import datetime
import os
import time
from typing import Literal
import pandas as pd
import tabulate
from wiliot_deployment_tools.ag.ut_defines import BRIDGE_ID, NFPKT, PAYLOAD, RSSI
from wiliot_deployment_tools.api.extended_api import GatewayType
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.if_defines import BRIDGES, DEFAULT_DELAY, COUPLING_DUPLICATIONS, MAX_NFPKT, MAX_RSSI, SEP, COUPLING_TIME_DELAYS
from wiliot_deployment_tools.interface.uart_if import UARTInterface
from wiliot_deployment_tools.gw_certificate.tests.static.coupling_defines import *
from wiliot_deployment_tools.interface.mqtt import MqttClient
from wiliot_deployment_tools.interface.pkt_generator import BrgPktGenerator, BrgPktGeneratorNetwork, TagPktGenerator
from wiliot_deployment_tools.gw_certificate.tests.static.generated_packet_table import TEST_COUPLING, CouplingRunData
from wiliot_deployment_tools.gw_certificate.tests.generic import PASS_STATUS, GenericTest, GenericStage
from wiliot_deployment_tools.interface.packet_error import PacketError


# TEST STAGES

class CouplingTestError(Exception):
    pass

class GenericCouplingStage(GenericStage):
    def __init__(self, mqttc:MqttClient, ble_sim:BLESimulator, gw_capabilities:GWCapabilities, 
                 stage_name, pass_criteria:Literal['all', '90%', 'none'], **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(stage_name=stage_name, **self.__dict__)
        
        #Clients
        self.mqttc = mqttc
        self.ble_sim = ble_sim

        # Packets list
        self.local_pkts = []
        self.mqtt_pkts = []
        
        # GW Capabilities
        self.gw_capabilities = gw_capabilities
        
        # Packet Error / Run data
        self.packet_error = PacketError()
        self.run_data = CouplingRunData
    
    def prepare_stage(self):
        super().prepare_stage()
        self.mqttc.flush_messages()
        self.ble_sim.set_sim_mode(True)
        
    def fetch_mqtt_from_stage(self):
        def process_payload(packet:dict):
            payload = packet[PAYLOAD]
            payload = payload.upper()
            if len(payload) == 62:
                if payload[:4] == '1E16':
                    payload = payload [4:]
            # big2little endian
            if payload[:4] == 'FCC6':
                payload = 'C6FC' + payload[4:]
            packet[PAYLOAD] = payload
            return packet
        mqtt_pkts = self.mqttc.get_coupled_tags_pkts()
        self.mqtt_pkts = list(map(lambda p: process_payload(p), mqtt_pkts))

    def compare_local_mqtt(self):
        self.fetch_mqtt_from_stage()
        local_pkts_df = pd.DataFrame(self.local_pkts)
        mqtt_pkts_df = pd.DataFrame(self.mqtt_pkts)
        if not set(SHARED_COLUMNS) <= set(mqtt_pkts_df.columns):
            missing_columns = list(set(SHARED_COLUMNS) - set(mqtt_pkts_df.columns))
            for missing_column in missing_columns:
                if missing_column in OBJECT_COLUMNS:
                    mqtt_pkts_df[missing_column] = ''
                if missing_column in INT64_COLUMNS:
                    mqtt_pkts_df[missing_column] = 0
        comparison = local_pkts_df
        received_pkts_df = pd.merge(local_pkts_df[SHARED_COLUMNS], mqtt_pkts_df[SHARED_COLUMNS], how='inner')
        received_pkts = set(received_pkts_df[PAYLOAD])
        comparison[RECEIVED] = comparison[PAYLOAD].isin(received_pkts)
        comparison['pkt_id'] = comparison['payload'].apply(lambda x: x[-8:])
        self.comparison = comparison
                
    def generate_stage_report(self):
        self.compare_local_mqtt()
        report = []
        num_pkts_sent = len(self.comparison)
        num_pkts_received = self.comparison['received'].eq(True).sum()
        self.stage_pass = num_pkts_sent == num_pkts_received
        report.append((('Number of coupled packets sent'), num_pkts_sent))
        report.append((('Number of coupled packets received'), num_pkts_received))
        self.add_to_stage_report(f'---Stage {self.stage_name} {PASS_STATUS[self.stage_pass]}, Running time {datetime.datetime.now() - self.start_time}')
        self.add_to_stage_report(tabulate.tabulate(pd.DataFrame(report), showindex=False))
        not_received = self.comparison[self.comparison[RECEIVED]==False][REPORT_COLUMNS]
        if len(not_received)>0:
            self.add_to_stage_report('Packets not received:')
            self.add_to_stage_report(tabulate.tabulate(not_received, headers='keys', showindex=False))
        self.comparison.to_csv(self.csv_path)
        self.add_to_stage_report(f'Stage data saved - {self.csv_path}')
        debug_print(self.report)
        return self.report
    
class InitStage(GenericCouplingStage):
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.pkt_gen = BrgPktGenerator()
    
    def run(self):
        self.start_time = datetime.datetime.now()
        for idx, duplication in enumerate(INIT_STAGES_DUPLICATIONS):
            new_pkt = self.pkt_gen.get_new_data_si()
            # First 2 runs generate smallest RSSI/NFPKT
            # Last 2 runs generate biggest RSSI/NFPKT
            if idx < 2:
                self.pkt_gen.set_rssi_nfpkt(rssi=idx, nfpkt=idx)
            else:
                self.pkt_gen.set_rssi_nfpkt(rssi=MAX_RSSI-idx+2, nfpkt=MAX_NFPKT-idx+2)
            new_pkt = self.pkt_gen.get_existing_data_si()
            expected_pkt = self.pkt_gen.get_expected_coupled_mqtt()
            data = new_pkt['data_packet']
            si = new_pkt['si_packet']
            expected_pkt.update({'duplication': duplication, 'time_delay': DEFAULT_DELAY,
            'si_rawpacket': si, 'data_rawpacket': data})
            self.local_pkts.append(expected_pkt)
            self.ble_sim.send_data_si_pair(data, si, duplication, delay=DEFAULT_DELAY)
        time.sleep(5)

class OneBrgStage(GenericCouplingStage):
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.pkt_gen = BrgPktGenerator()
    
    def run(self):
        self.start_time = datetime.datetime.now()
        for duplication in COUPLING_DUPLICATIONS: #tqdm(duplications, desc='Duplications', position=1, leave=True):
            debug_print(f'Duplication {duplication}')
            for time_delay in COUPLING_TIME_DELAYS: #tqdm(time_delays, desc='Time Delays', position=2, leave=True):
                debug_print(f'Time Delay {time_delay}')
                run_data = self.run_data.get_data(TEST_COUPLING, duplication, time_delay, BRIDGES[0])
                data = run_data.data
                si = run_data.si
                packet_error = run_data.packet_error
                self.local_pkts.append(run_data.expected_mqtt)
                self.ble_sim.send_data_si_pair(data, si, duplication, delay=time_delay, packet_error=packet_error)
        time.sleep(5)

class ThreeBrgInitStage(GenericCouplingStage):
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.brg_network = BrgPktGeneratorNetwork()
    
    def run(self):
        self.start_time = datetime.datetime.now()
        duplication = COUPLING_DUPLICATIONS[3]
        time_delay = COUPLING_TIME_DELAYS[2]
        # Construct packet list from data
        pkts = []
        for brg_idx in BRIDGES:
            pkt = {}
            run_data = self.run_data.get_data(TEST_COUPLING, duplication, time_delay, brg_idx)
            pkt['data_packet'] = run_data.data
            pkt['si_packet'] = run_data.si
            pkt['time_delay'] = run_data.scattered_time_delay
            pkt['packet_error'] = run_data.packet_error
            pkt['bridge_id'] = run_data.bridge_id
            self.local_pkts.append(run_data.expected_mqtt)
            pkts.append(pkt)
        # Send scattered packets
        self.ble_sim.send_brg_network_pkts(pkts, duplication)
        time.sleep(5)
        


class ThreeBrgStage(GenericCouplingStage):
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.brg_network = BrgPktGeneratorNetwork()

    
    def run(self):
        self.start_time = datetime.datetime.now()
        for duplication in COUPLING_DUPLICATIONS:
            debug_print(f'Duplication {duplication}')
            # Time delays from 45 -> 255
            for time_delay in COUPLING_TIME_DELAYS[1:]:
                debug_print(f'Time Delay {time_delay}')
                # Construct packet list from data
                pkts = []
                for brg_idx in BRIDGES:
                    pkt = {}
                    run_data = self.run_data.get_data(TEST_COUPLING, duplication, time_delay, brg_idx)
                    pkt['data_packet'] = run_data.data
                    pkt['si_packet'] = run_data.si
                    pkt['time_delay'] = run_data.scattered_time_delay
                    pkt['packet_error'] = run_data.packet_error
                    pkt['bridge_id'] = run_data.bridge_id
                    self.local_pkts.append(run_data.expected_mqtt)
                    pkts.append(pkt)
                # Send scattered packets
                self.ble_sim.send_brg_network_pkts(pkts, duplication)
        time.sleep(5)

class IncrementalStage(GenericCouplingStage):
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.pkt_gen = TagPktGenerator(adva=INCREMENTAL_STAGE_ADVA)
    
    def run(self):
        self.start_time = datetime.datetime.now()
        for time_delay in INCREMENTAL_TIME_DELAYS:
            for pkt in INCREMENTAL_PACKETS:
                data = self.pkt_gen.get_packet()
                # expected_pkt = self.pkt_gen.get_expected_mqtt()
                # expected_pkt.update({'duplication': duplication, 'time_delay': DEFAULT_DELAY,
                # 'si_rawpacket': si, 'data_rawpacket': data})
                # self.local_pkts.append(expected_pkt)
                self.ble_sim.send_packet(raw_packet=data, duplicates=1, delay=time_delay)
                self.pkt_gen.set_pkt_id(self.pkt_gen.get_pkt_id() + 1)
        time.sleep(5)

#TODO - Add APIValidation Stage

# TEST CLASS

STAGES = [InitStage, OneBrgStage, ThreeBrgInitStage, ThreeBrgStage]

class CouplingTest(GenericTest):
    def __init__(self, **kwargs):        
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, test_name=type(self).__name__)
        self.all_messages_in_test = []
        self.stages = [stage(**self.__dict__) for stage in STAGES]

    def run(self):
        super().run()
        self.enter_dev_mode()
        self.test_pass = True
        for stage in self.stages:
            stage.prepare_stage()
            stage.run()
            self.add_to_test_report(stage.generate_stage_report())
            if stage.stage_pass == False:
                self.test_pass = False
        self.end_test()
        self.exit_dev_mode()
    
# TODO - make sure everything is running non-randomized



import binascii
import datetime
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import pandas as pd
import numpy as np
import pkg_resources
import os
import subprocess
import shutil
from wiliot_deployment_tools.gw_certificate.tests.static.coupling_defines import *
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.pkt_generator import TagPktGenerator
from wiliot_deployment_tools.interface.uart_if import UARTInterface

if __name__ == "__main_____":
    test_name = 'zebra_screenoff'
    CSV_NAME = test_name
    CSV_PATH = pkg_resources.resource_filename(__name__, f'{test_name}/{CSV_NAME}.csv')
    ANDROID_LOG_PATH = pkg_resources.resource_filename(__name__, f'{test_name}/{CSV_NAME}.log')
    HTML_PATH_BAR = pkg_resources.resource_filename(__name__, f'{test_name}/{test_name}_bar.html')
    HTML_PATH_SCATTER = pkg_resources.resource_filename(__name__, f'{test_name}/{test_name}_scatter.html')
    
    os.makedirs(pkg_resources.resource_filename(__name__, f'{test_name}'), exist_ok=True)
    
    subprocess.run('adb -d logcat -c', shell=True)
    subprocess.Popen(f'adb -d logcat --format="epoch" > {ANDROID_LOG_PATH}', shell=True)
    pkts = []
    
    pkt_gen = TagPktGenerator(adva=INCREMENTAL_STAGE_ADVA)
    uart = UARTInterface('/dev/cu.usbserial-0001')
    ble_sim = BLESimulator(uart)
    ble_sim.set_sim_mode(True)
    for time_delay in INCREMENTAL_TIME_DELAYS:
        for pkt in INCREMENTAL_PACKETS:
            data = pkt_gen.get_packet()
            # expected_pkt = self.pkt_gen.get_expected_mqtt()
            # expected_pkt.update({'duplication': duplication, 'time_delay': DEFAULT_DELAY,
            # 'si_rawpacket': si, 'data_rawpacket': data})
            # self.local_pkts.append(expected_pkt)
            ble_sim.send_packet(raw_packet=data, duplicates=1, delay=15)
            pkt_id_bytes, pkt_id_int = pkt_gen.get_pkt_id()
            pkt_gen.set_pkt_id(pkt_id_int + 1)
            pkts.append({PAYLOAD: data, 'time_delay':time_delay,
                        'time': datetime.datetime.now(), 'source':'ble',
                        'rssi': 0})
    pkts = pd.DataFrame().from_dict(pkts)
    pkts.to_csv(CSV_PATH)
            
    # Parse android logs
    android_data = []
    with open(ANDROID_LOG_PATH) as f:
        loglines = f.readlines()
        
    for line in loglines:
        if 'AdvA' not in line:
            continue
        line = line.split()
        time = datetime.datetime.fromtimestamp(float(line[0]))
        adva = line[7]
        rssi = float(line[8].split(':')[-1])
        payload = line[9].split('[')[-1]+':'+line[11].split(']')[0]
        android_data.append({'source': 'android', 'time': time,
                                'adva': adva, 'rssi': rssi, 'payload': payload})
    android_data = pd.DataFrame().from_dict(android_data)
    
    all_data = pd.concat([pkts, android_data])
    all_data['pkt_counter'] = all_data['payload'].apply(lambda x: int.from_bytes(binascii.unhexlify(str(x)[-4:]), 'big'))
    percentages = []
    for delay in INCREMENTAL_TIME_DELAYS:
        relevant_pkt_ids = all_data[(all_data['source']=='ble') & (all_data['time_delay']==delay)]['pkt_counter']
        num_sent = len(relevant_pkt_ids.unique())
        num_received = len(all_data[(all_data['source']=='android') & (all_data['pkt_counter'].isin(relevant_pkt_ids))]['payload'].unique())
        percentages.append({'time_delay': delay, 'source': 'ble', 'num_pkts':num_sent, 'percent': num_sent/num_sent})
        percentages.append({'time_delay': delay, 'source': 'android', 'num_pkts':num_received, 'percent': num_received/num_sent})
    percentages = pd.DataFrame().from_dict(percentages)
    trace1 = px.scatter(all_data, x='time', y='pkt_counter', render_mode='svg', color='source', hover_data=['rssi'], title=test_name)
    trace2 = px.bar(percentages, x='time_delay', y='num_pkts', color='source', barmode='group', hover_data=['percent'])
    trace2.update_xaxes(type='category')

    trace1.write_html(HTML_PATH_SCATTER)
    trace2.write_html(HTML_PATH_BAR)
    # fig.show()
    # fig.write_html(HTML_PATH)


from argparse import ArgumentParser
from wiliot_deployment_tools.gw_certificate.gw_certificate import GWCertificate

def main():
    parser = ArgumentParser(prog='wlt-gw-certificate',
                            description='Gateway Certificate - CLI Tool to test Wiliot GWs')
    required = parser.add_argument_group('required arguments')
    required.add_argument('-owner', type=str, help="Owner ID", required=True)
    required.add_argument('-gw', type=str, help="Gateway ID", required=True)
    
    optional = parser.add_argument_group('optional arguments')
    optional.add_argument('-suffix', type=str, help="Topic suffix", default='', required=False)
    optional.add_argument('-coupled', action='store_true', help="GW Metadata Coupling Enabled", default='False', required=False)

    
    args = parser.parse_args()
    topic_suffix = '' if args.suffix == '' else '-'+args.suffix
    gwc = GWCertificate(gw_id=args.gw, owner_id=args.owner, coupled=args.coupled,topic_suffix=topic_suffix)
    gwc.run_tests()
    gwc.create_results_html()

def main_cli():
    main()

if __name__ == '__main__':
    main()



INIT_STAGE = "InitStage"
EXTENDED_STAGE = "ExtendedStage"
STAGE_CONFIGS = {
    INIT_STAGE:([i for i in range(700, 901, 100)], range(3)),
    EXTENDED_STAGE:([i for i in range(100, 2001, 100)], range(5))}
TX_MAX_DURATIONS = range(100, 501, 100)
RETRIES = range(5)
MAX_RX_TX_PERIOD_SECS = 0.255
DEFAULT_BRG_ID = "FFFFFFFFFFFF"


import binascii
import datetime
from typing import Literal, List
import numpy as np
import random
import os

from wiliot_deployment_tools.ag.ut_defines import (
    BRIDGE_ID,
    NFPKT,
    PAYLOAD,
    RSSI,
    TIMESTAMP,
)
from wiliot_deployment_tools.interface.if_defines import ADVA_LENGTH, GAP_LENGTH, GROUP_ID_LENGTH, SERVICE_UUID_LENGTH
from wiliot_deployment_tools.ag.wlt_types import *
from wiliot_deployment_tools.ag.wlt_types_ag import (
    GROUP_ID_SIDE_INFO,
    GROUP_ID_BRG2GW,
    GROUP_ID_GW2BRG,
    Hdr,
    SideInfo,
)
from wiliot_deployment_tools.ag.wlt_types_data import DataPacket

# Shared Functions
def generate_random_adva(device:Literal['Tag', 'Bridge']):
    """generate random 6-byte AdvA

    :return: AdvA (12 hex chars)
    :rtype: str
    """
    adva = os.urandom(6).hex().upper()
    if device == 'Tag':
        adva = apply_adva_bitmask(adva, 'non_resolvable')
    if device == 'Bridge':
        adva = apply_adva_bitmask(adva, 'random_static')
    return adva

def apply_adva_bitmask(adva, address_type:Literal['random_static', 'non_resolvable']):
    """apply AdvA Bitmask 
    BRGs AdvA comply with BTLE Random Static Address
    Pixel AdvA comply with BTLE Random Private Non-Resolvable Address

    :param adva: AdvA
    :type adva: str
    :return: AdvA (12 hex chars)
    :rtype: str
    """ 
    masked_adva = bytearray(binascii.unhexlify(adva))
    if address_type == 'random_static':
        masked_adva[5] = masked_adva[5] | 0b11000000
    if address_type == 'non_resolvable':
        masked_adva[5] = masked_adva[5] & 0b00111111
    masked_adva = masked_adva.hex().upper()
    return masked_adva

def init_adva(adva=None, device:Literal['Tag', 'Bridge']='Tag'):
    """Init function helper

    :param adva: AdvA, defaults to None
    :type adva: str, optional
    :return: AdvA (12 hex chars)
    :rtype: str
    """
    if adva is None:
        adva = generate_random_adva(device)
    adva = apply_adva_bitmask(adva, device)
    return adva

def increment_circular(x, num_to_inc, bits):
    """Increment object of {bits} size circularly
    :param x: number to increment
    :type x: int
    :param num_to_inc: number to increment
    :type num_to_inc: int
    :param bits: bit size
    :type bits: int
    """
    return (x + num_to_inc) % (2 ^ bits)

def decrease_circular(x, num_to_dec, bits):
    """Decrease object of {bits} size circularly
    :param x: number to decrease
    :type x: int
    :param num_to_dec: number to decrease
    :type num_to_dec: int
    :param bits: bit size
    :type bits: int
    """
    return (x-num_to_dec) % (2 ^ bits)

def generate_random_bridge_id():
    """Generate random 6 byte bridge ID

    :return: bridge ID - (12 hex chars, int)
    :rtype: tuple (str, int)
    """
    bridge_id = os.urandom(6).hex().upper()
    bridge_id_int = int.from_bytes(binascii.unhexlify(bridge_id), "big")
    return bridge_id, bridge_id_int


class TagPktGenerator:
    """Tag Packet Generator - represents 1 Wiliot Pixel"""
    def __init__(self, adva=None, group_id=None, service_uuid:Literal["tag", "bridge"]='bridge') -> None:
        """
        :param adva: AdvA, defaults to None
        :type adva: str, optional
        :param group_id: Tag's group id, defaults to 0x020000 if none given
        :type group_id: int, optional
        :param service_uuid: set the generator's service uuid ("tag"=0xAFFD, "bridge"=0xC6FC)
        :type service_uuid: string
        """
        self.adva = init_adva(adva, device='Tag')
        self.data_packet = DataPacket(group_id=group_id, service_uuid=service_uuid)
        pass
    
    def __repr__(self) -> str:
        return f'Tag: {self.adva}, PktID: {self.get_pkt_id()[0]}'
    
    def get_pkt_id(self):
        """Get generator's current packet ID

        :return: 6 byte packet ID - (12 hex chars, int)
        :rtype: tuple (str, int)
        """
        pkt_id_int = self.data_packet.generic.pkt_id
        pkt_id_bytes = pkt_id_int.to_bytes(4, "big").hex()
        return pkt_id_bytes, pkt_id_int
    
    def set_pkt_id(self, pkt_id_int):
        """Set generator's packet ID

        :param pkt_id_int: packet ID
        :type pkt_id_int: int
        """
        assert 0 < pkt_id_int < (2 ** 32), "PacketID Must be a 32 bit unsigned integer!"
        self.data_packet.generic.pkt_id = pkt_id_int
        
    def randomize_pkt_id(self):
        """Randomize generator's packet ID"""
        pkt_id_int = int.from_bytes(os.urandom(4), "big")
        self.set_pkt_id(pkt_id_int)
    
    def randomize_packet_payload(self):
        """Randomize Generator's packet payload (keep packet ID as is)"""
        self.data_packet.generic.payload = int.from_bytes(os.urandom(20), "big")
    
    def increment_pkt_data(self):
        packet = self.get_packet()
        data = packet[(ADVA_LENGTH+GAP_LENGTH+SERVICE_UUID_LENGTH+GROUP_ID_LENGTH):]
        incremented_data = [bytes([((int(data[i:i+2], 16) + 1) % 256)]).hex().upper() for i in range(0, len(data), 2)]
        incremented_packet = packet[:ADVA_LENGTH+GAP_LENGTH+SERVICE_UUID_LENGTH+GROUP_ID_LENGTH] + ''.join(incremented_data)
        self.set_packet(incremented_packet)

    def get_packet(self):
        """Get current packet from generator (hex string)

        :return: Tag's Packet [12 char AdvA + 62 char BLE Packet]
        :rtype: str
        """
        return self.adva + self.data_packet.dump()
    
    def set_packet(self, pkt):
        """set packet to input pkt"""
        if len(pkt) == 74:
            self.adva = pkt[:12]
            pkt = pkt[12:]
        assert len(pkt) == 62, "packet must be 74 / 62 hex chars long!"
        self.data_packet.set(pkt)
    
    def get_expected_mqtt(self):
        timestamp = int(datetime.datetime.now().timestamp() * 1000)
        expected = {
            TIMESTAMP: timestamp,
            PAYLOAD: self.get_packet()[ADVA_LENGTH+GAP_LENGTH:]
        }
        return expected



class BrgPktGenerator:
    """Bridge Packet Generator - represents 1 wiliot Bridge"""
    def __init__(self, bridge_id:str=None):
        """
        :param bridge_id: bridge ID, defaults to randomly generated if None
        :type bridge_id: str, optional
        """
        if bridge_id is None:
            self.bridge_id, self.bridge_id_int = generate_random_bridge_id()
        else:
            self.bridge_id = bridge_id
            self.bridge_id_int = int.from_bytes(binascii.unhexlify(bridge_id), "big")
        self.adva = apply_adva_bitmask(self.bridge_id, 'random_static')
        # Data packet init
        self.tag_list:List[TagPktGenerator] = []
        self.append_data_pkt()
        # SI packet init
        self.si_list:List[WltPkt]
        self.update_si_list()
        
        # BRG CFG packet init
        self.brg_cfg = WltPkt(
            hdr=Hdr(group_id=GROUP_ID_BRG2GW),
            generic=Brg2GwCfgV7(
                msg_type=BRG_MGMT_MSG_TYPE_CFG_INFO,
                major_ver=3,
                minor_ver=12,
                build_ver=35,
                brg_mac=self.bridge_id_int,
            ),
        )
        self.brg_hb = WltPkt(
            hdr=Hdr(group_id=GROUP_ID_BRG2GW),
            generic=Brg2GwHbV7(
                msg_type=BRG_MGMT_MSG_TYPE_HB, brg_mac=self.bridge_id_int
            ),
        )
        self.brg_seq_id = 0
        
    def __repr__(self):
        return f'BRG {self.bridge_id}, Tags {self.tag_list}'

    def set_bridge_id(self, brg_mac):
        """Set generator's bridge ID

        :param brg_mac: bridge ID (12 hex chars)
        :type brg_mac: str
        """
        self.bridge_id = brg_mac
        self.bridge_id_int = int.from_bytes(binascii.unhexlify(brg_mac), "big")
        self.brg_cfg = WltPkt(
            hdr=Hdr(group_id=GROUP_ID_BRG2GW),
            generic=Brg2GwCfgV7(
                msg_type=BRG_MGMT_MSG_TYPE_CFG_INFO,
                major_ver=3,
                minor_ver=12,
                build_ver=35,
                brg_mac=self.bridge_id_int,
            ),
        )
        self.brg_hb = WltPkt(
            hdr=Hdr(group_id=GROUP_ID_BRG2GW),
            generic=Brg2GwHbV7(
                msg_type=BRG_MGMT_MSG_TYPE_HB, brg_mac=self.bridge_id_int
            ),
        )
    
    def set_random_bridge_id(self):
        bridge_id, bridge_id_int = generate_random_bridge_id()
        self.set_bridge_id(bridge_id)
    
    # Data Packet
    
    def append_data_pkt(self, data_pkt=None):
        """Append data packet to tag list
        if no data packet is input, a new TagPktGenerator object will be appended to tag list

        :param data_pkt: 62 hex char string, defaults to None
        :type data_pkt: str, optional
        """
        idx_to_append = len(self.tag_list)
        self.tag_list.append(None)
        self.set_data_pkt(tag_idx=idx_to_append, data_pkt=data_pkt)

    def set_data_pkt(self, tag_idx, data_pkt=None):
        """Set tag generator at specific idx's data packet from string
        if no data packet is input, a new TagPktGenerator object will be created at tag_idx

        :param tag_idx: tag index
        :type tag_idx: int
        :param data_pkt: data packet (62 hex chars), defaults to None
        :type data_pkt: str, optional
        """
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        tag_pkt_gen = TagPktGenerator()
        if data_pkt is not None:
            tag_pkt_gen.set_packet(data_pkt)
        self.tag_list[tag_idx] = tag_pkt_gen
    
    # SideInfo Packet
    
    def generate_si_from_pkt_id(self, pkt_id_int):
        """Generate SI from packet ID

        :param pkt_id_int: packet ID
        :type pkt_id_int: int
        :return: WltPkt for SideInfo, with RSSI, NFPKT zeroed out
        :rtype: WltPkt(hdr:Hdr, generic:SideInfo)
        """
        assert 0 <= pkt_id_int < (2 ** 32), "Packet ID must be a 32 bit unsigned integer!"
        si_packet = WltPkt(
            hdr=Hdr(group_id=GROUP_ID_SIDE_INFO),
            generic=SideInfo(brg_mac=self.bridge_id_int, pkt_id=pkt_id_int),
        )
        return si_packet
    
    def update_si_list(self):
        """Update the SI list according to current pixel's packet IDs"""
        si_table = []
        for tag in self.tag_list:
            tag_pkt_id_bytes, tag_pkt_id_int = tag.get_pkt_id()
            si_table.append(self.generate_si_from_pkt_id(tag_pkt_id_int))
        self.si_list = si_table
        
    def randomize_si_rssi_nfpkt(self):
        """Randomize all SI's rssi and nfpkt values
        """
        for si in self.si_list:
            rssi = int.from_bytes(os.urandom(1), "big")
            nfpkt = int.from_bytes(os.urandom(2), "big")
            si.generic.rssi = rssi
            si.generic.nfpkt = nfpkt

    def set_rssi_nfpkt(self, rssi, nfpkt, tag_idx:int=None):
        if tag_idx is None:
            tag_idx = 0
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        self.si_list[tag_idx].generic.rssi = rssi
        self.si_list[tag_idx].generic.nfpkt = nfpkt
    
    # Coupling (Shared Data + SideInfo)
    
    def set_tag_list(self, tag_list:List[TagPktGenerator]):
        self.tag_list = tag_list
        self.update_si_list()
    
    def increment_pkt_id(self, tag_idx:int=None):
        """Increment tag generator at specific idx's data packet's packet ID 

        :param tag_idx: tag index, defaults to None
        :type tag_idx: int, optional
        """
        if tag_idx is None:
            tag_idx = 0
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        self.tag_list[tag_idx].generic.pkt_id = increment_circular(self.tag_list[tag_idx].generic.pkt_id, 1, 32)
        self.update_si_list()

    def randomize_data_packet(self, tag_idx:int=None):
        """Randomize tag generator at specific idx's data packet (payload + ID),
        Update SI to match

        :param tag_idx: tag index, defaults to None
        :type tag_idx: int, optional
        """
        if tag_idx is None:
            tag_idx = 0
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        tag = self.tag_list[tag_idx]
        # randomize packet ID
        tag.randomize_pkt_id()
        # randomize packet payload
        tag.randomize_packet_payload()
        # update SI
        self.update_si_list()


    def get_data(self, tag_idx:int):
        """Get tag at index's data packet (AdvA + BLE Packet)

        :param tag_idx: tag index
        :type tag_idx: int
        :return: data packet (74 hex char)
        :rtype: str
        """
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        return self.tag_list[tag_idx].get_packet()
    
    def get_si(self, tag_idx:int):
        """Get tag at index's sideinfo packet (AdvA + BLE Packet)

        :param tag_idx: _description_
        :type tag_idx: int
        :return: _description_
        :rtype: _type_
        """
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        return self.adva + self.si_list[tag_idx].dump()

    def get_expected_coupled_mqtt(self, tag_idx:int=None):
        if tag_idx is None:
            tag_idx = 0
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        si = self.si_list[tag_idx]
        """generates expected MQTT packet"""
        timestamp = int(datetime.datetime.now().timestamp() * 1000)
        expected = {
            TIMESTAMP: timestamp,
            BRIDGE_ID: self.bridge_id,
            NFPKT: si.generic.nfpkt,
            RSSI: si.generic.rssi,
            PAYLOAD: self.get_data(tag_idx)[(ADVA_LENGTH+GAP_LENGTH):]
        }
        return expected
    
    def get_expected_uncoupled_mqtt(self, tag_idx:int=None):
        if tag_idx is None:
            tag_idx = 0
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        timestamp = int(datetime.datetime.now().timestamp() * 1000)
        expected_data = {
            TIMESTAMP: timestamp,
            PAYLOAD: self.get_data(tag_idx)[(ADVA_LENGTH+GAP_LENGTH):]
        }
        expected_si = {
            TIMESTAMP: timestamp,
            PAYLOAD: self.get_si(tag_idx)[(ADVA_LENGTH+GAP_LENGTH):]
        }
        return [expected_data, expected_si]
        
    def get_expected_hb_mqtt(self):
        timestamp = int(datetime.datetime.now().timestamp() * 1000)
        expected = {
            TIMESTAMP: timestamp,
            PAYLOAD: self.get_brg_hb()[(ADVA_LENGTH+GAP_LENGTH):]
        }
        return expected
    # Management packets (Heartbeat / CFG)

    def increment_brg_seq_id(self):
        """Increment Bridge sequenceID (Relevant for CFG/HB Packets only)"""
        self.brg_seq_id = increment_circular(self.brg_seq_id, 2, 8)
        self.brg_cfg.generic.seq_id = decrease_circular(self.brg_seq_id, 2, 8)
        self.brg_hb.generic.seq_id = decrease_circular(self.brg_seq_id, 1, 8)

    def increment_hb_counters(self, num=1):
        """Update bridge HB counters values

        :param num: num to increment by, defaults to 1
        :type num: int, optional
        """
        self.brg_hb.generic.non_wlt_rx_pkts_ctr = increment_circular(self.brg_hb.generic.non_wlt_rx_pkts_ctr, num, 24)
        self.brg_hb.generic.bad_crc_pkts_ctr = increment_circular(self.brg_hb.generic.bad_crc_pkts_ctr, num, 24)
        self.brg_hb.generic.wlt_rx_pkts_ctr = increment_circular(self.brg_hb.generic.wlt_rx_pkts_ctr, num, 24)
        self.brg_hb.generic.wlt_tx_pkts_ctr = increment_circular(self.brg_hb.generic.wlt_tx_pkts_ctr, num, 16)
        self.brg_hb.generic.tags_ctr = len(self.tag_list)

    def increment_all(self):
        """Increment Bridge sequenceID and HB counters"""
        self.increment_brg_seq_id()
        self.increment_hb_counters()
        
    def get_brg_cfg(self):
        """Get bridge CFG Packet

        :return: CFG Packet (74 hex chars)
        :rtype: str
        """
        return self.adva + self.brg_cfg.dump()

    def get_brg_hb(self):
        """Get bridge HB Packet

        :return: HB Packet (74 hex chars)
        :rtype: str
        """
        return self.adva + self.brg_hb.dump()

    # General

    def get_existing_data_si(self, tag_idx:int=None) -> dict:
        """Get tag at tag_idx's Data packet and relevant SideInfo packet

        :param tag_idx: tag index, defaults to None
        :type tag_idx: int, optional
        :return: dictionary with data packet and sideinfo packet
        :rtype: dict
        """
        if tag_idx is None:
            tag_idx = 0
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        existing_packets = {
            "data_packet": self.get_data(tag_idx),
            "si_packet": self.get_si(tag_idx)
        }
        return existing_packets
    
    def get_new_data_si(self, tag_idx:int=None) -> dict:
        """Generate new Data+SideInfo packets for tag at tag_idx

        :param tag_idx: tag index, defaults to None
        :type tag_idx: int, optional
        :return: dictionary with data packet and sideinfo packet
        :rtype: dict
        """
        if tag_idx is None:
            tag_idx = 0
        assert tag_idx <= len(self.tag_list), f'Tag index must be in {[i for i in range(len(self.tag_list))]}'
        self.randomize_data_packet(tag_idx)
        self.randomize_si_rssi_nfpkt()
        self.increment_brg_seq_id()
        self.increment_hb_counters()
        return self.get_existing_data_si(tag_idx)

class BrgPktGeneratorNetwork:
    """Bridge Packet Generator Network - Represents multiple bridges setup at same location and echoing the same tags"""
    def __init__(self, num_brgs=3):
        """
        :param num_brgs: number of bridges in array, defaults to 3
        :type num_brgs: int, optional
        """
        assert num_brgs > 1, 'BrgArray cannot be smaller than 1!'
        self.brg_list = [BrgPktGenerator() for brg in range(num_brgs)]
        self.primary_brg = self.brg_list[0]
        self.secondary_brgs = self.brg_list[1:]
        for brg in self.secondary_brgs:
            brg.set_tag_list(self.primary_brg.tag_list)
    
    def get_new_pkt_pairs(self) -> list:
        """Get data/si pairs for all bridges in bridge network

        :return: List of dictionaries containing {bridge_id, data_packet, si_packet} key:value pairs
        :rtype: list[dict]
        """
        pkts = []
        new_pkt_primary = self.primary_brg.get_new_data_si()
        pkts.append(new_pkt_primary)
        for brg in self.secondary_brgs:
            brg.update_si_list()
            brg.randomize_si_rssi_nfpkt()
            pkts.append(brg.get_existing_data_si())
        for idx, pkt in enumerate(pkts):
            pkt.update({'bridge_id': self.brg_list[idx].bridge_id})
        return pkts
    
    def get_scattered_pkt_pairs(self, delay):
        """Get data/si pairs with scattered time delays

        :param delay: bridges rxTxPeriod equivalent (millisecond)
        :type delay: int
        """
        assert delay / len(self.brg_list) >= 10, 'Cannot get scattered PKT pairs for parameters! decrease num of brgs or increase delay'
        start_times = sorted(random.sample(list(np.arange(0, delay-10+1, 10)), k=len(self.brg_list)))
        start_times.append((start_times[0] + delay))
        pkts = self.get_new_pkt_pairs()
        
        for idx, pkt in enumerate(pkts):
            # minimum time between UART packet commands = 10ms
            pkt['time_delay'] = max(start_times[idx+1] - start_times[idx], 10)
        return pkts

class CloudPktGenerator:
    """Cloud Packet Generator - Represents """
    def __init__(self):
        self.bridge_id, self.bridge_id_int = generate_random_bridge_id()
        pass

    def generate_config_pkt(self, seq_id=None, bridge_id=None):
        if seq_id == None:
            seq_id = int.from_bytes(os.urandom(1), 'big')
        if bridge_id is not None:
            bridge_id_int = int.from_bytes(binascii.unhexlify(bridge_id), "big")

        pkt = WltPkt(Hdr(group_id=GROUP_ID_GW2BRG), generic=Gw2BrgCfgV7(
            msg_type=BRG_MGMT_MSG_TYPE_CFG_SET,
            global_pacing_group=random.getrandbits(4),
            output_power_sub_1_ghz=random.getrandbits(4),
            seq_id=random.getrandbits(8),
            brg_mac=bridge_id_int,
            unused0=0,
            pkt_types_mask=random.getrandbits(5),
            unused1=0,
            rxtx_period=random.getrandbits(8),
            tx_period=random.getrandbits(8),
            energy_pattern_idx=random.getrandbits(8),
            output_power_2_4=random.getrandbits(8),
            pacer_interval=random.getrandbits(16),
            unused2=0,
            tx_prob=random.getrandbits(3),
            tx_repetition=random.getrandbits(4),
            transmit_time_sub_1_ghz=random.getrandbits(4),
            sub1g_freq_profile=random.getrandbits(4),
            ))
        return pkt.dump()

    # def generate_action_pkt(self, action: Literal["blink", "reboot"], seq_id=None, bridge_id=None):
    #     if seq_id == None:
    #         seq_id = int.from_bytes(os.urandom(1), 'big')
    #     if bridge_id is not None:
    #         bridge_id_int = int.from_bytes(binascii.unhexlify(bridge_id), "big")

    #     pkt = WltPkt(Hdr(group_id=GROUP_ID_GW2BRG), generic=ActionV7(
    #         msg_type=BRG_MGMT_MSG_TYPE_ACTION,
    #         seq_id=seq_id,
    #         action_id=
    #                                                                  )


import unittest
from wiliot_deployment_tools.common.debug import *


class TestDebuggingUtils(unittest.TestCase):

    def test_print_git_info(self):
        self.assertTrue(print_package_gitinfo())
    
    def test_if_databricks(self):
        self.assertFalse(is_databricks())


if __name__ == '__main__':
    unittest.main()



from argparse import ArgumentParser
from wiliot_deployment_tools.power_optimization_tool.current_gain_analysis import OptimalPowerConfiguration

def main():
    parser = ArgumentParser(description="Calculate and plot Avg Current vs. Relative Gain")
    required = parser.add_argument_group('required arguments')
    required.add_argument("-bridge_type", type=str, help="Bridge type (minew or fanstel)", required=True)
    required.add_argument("-dc", nargs="+", type=float, help="The values of the duty cycles to compare. The order of insertion must match the order of the corresponding rxtx period", required=True)
    required.add_argument("-rxtx_period", nargs="+", type=int, help="The value of rxtx period. The order of insertion must match the order of the corresponding duty cycle", required=True)
    
    args = parser.parse_args()
    b = OptimalPowerConfiguration(args.bridge_type, args.dc, args.rxtx_period)
    b.generate_current_gain_plot()
    
def main_cli():
    main()

if __name__ == '__main__':
    main()


# Brownout / OTA Defines
SEP = "#" * 50
MINUTES_TO_BROWN_OUT = 5
TABLES_SYNC_MINUTES = 70
MINUTES_FOR_LC = 2
LATEST_FW = "1.9.0"
SINGLE_BAT_BO_DICT = {"2.4GhzOutputPower": 2, 'txPeriodMs': 75, 'rxTxPeriodMs': 255, 'energyPattern': 25}
DUAL_BAT_BO_DICT = {"2.4GhzOutputPower": 2, 'txPeriodMs': 75, 'rxTxPeriodMs': 255, 'energyPattern': 50}
LEGACY_BO_DICT = {'energyPattern': 17, 'rxTxPeriodMs': 99}
BO_DICT = {'energyPattern': 36}

# Configuration tool / change Brg/Gw Defines
ATC_GW_CONFIG = {
    'wifi': {'gwDataSrc': 'Bridges only (ch38)',
             'pacerInterval': 60,
             'txPeriodMs': 3,
             'rxTxPeriodMs': 90,
             'energizingPattern': 17,
             "2.4GhzOutputPower": 8},
    'lte': {},
    'mobile': {},
    'unknown': {}}
ATC_REGION_DICT = {
    'IL': {
        'sub1GhzFrequency': 919100,
        'energyPattern': 50
    }
}

energy_patterns = [18, 25, 26]
energy_patterns_db = {17: 50, 18: 51, 25: 56, 26: 57}
gw_rx_channel = {17: "Bridges only (ch39)", 18: "Bridges only (ch38)", 24: "Bridges only (ch39)",
                 25: "Bridges only (ch39)", 26: "Bridges only (ch39)", 51: "Bridges only (ch38)",
                 55: "Bridges only (ch39)", 56: "Bridges only (ch39)", 57: "Bridges only (ch39)",
                 50: "Bridges only (ch39)"}
shifted_brg_energy_patterns = [33, 34, 35, 58, 59, 60]
shifted_gw_energizing_patterns = [33]
lc_output_powers = [14, 17, 20, 23, 26, 29, 32]
fp_duty_cycles = [0.1, 0.15, 0.2, 0.25, 0.3]

# Power Mgmt
EXIT_POWER_MGMT_GW_DICT = {
    'gwDataSrc': gw_rx_channel[17],
    'txPeriodMs': 3,
    'rxTxPeriodMs': 15,
    'energizingPattern': 17,
    'gwMgmtMode': 'active'
}
KEEP_ALIVE_PERIOD = 30  # seconds
KEEP_ALIVE_SCAN_DURATION = 300  # in millisecond
SEC_TO_SEND = 2
BROADCAST_DST_MAC = 'FFFFFFFFFFFF'

colors = ['red', 'blue', 'yellow', 'cyan', 'green', 'brown', 'orange', 'pink', 'purple', 'black']

# Test Tool
WH_OWNER = '832742983939'
INIT_GW_CONFIG = {
    'wifi':{'gwDataSrc': gw_rx_channel[18],
            'gwDataMode': gw_rx_channel[18],
            'txPeriodMs': 3,
            'rxTxPeriodMs': 255,
            'energizingPattern': 17,
            'pacerInterval': 60},
    'lte':{},
    'mobile':{},
    'unknown':{}}
INIT_BRG_CONFIG = {
	'energy2400': {'config': {'txPeriod': 5, 'rxTxPeriod': 15, 'outputPower': 2, 'energyPattern': 36}},
	'datapath': {'config': {'txRepetition': 0, 'pacerInterval': 15, 'globalPacingGroup': 0}}, 
	'energySub1g': {'config': {'frequency': 919100, 'outputPower': 32}}
    }
	# 'otaUpgradeEnabled': False}

# prev init cfg for prev brgs version 
INIT_BRG_PREV_CONFIG = {
    "energyPattern": 36,
    "2.4GhzOutputPower": 2,
    "sub1GhzOutputPower": 32,
    "sub1GhzFrequency": 919100,
    "rxTxPeriodMs": 15,
    "txPeriodMs": 5,
    "txProbability": 50,
    "otaUpgradeEnabled": 0,
    "pacerInterval": 15,
    "globalPacingGroup": 0,
    "txRepetition": 0}

BRG_KEYS = ['adType0', 'adType1', 'uuidLsb0', 'uuidLsb1', 'uuidMsb0', 'uuidMsb1', 'sensor0Scramble', 'sensor1Scramble',
            'txPeriod', 'rxTxPeriod', '2.4GhzOutputPower', 'energyPattern',
            'pktTypesMask', 'txRepetition', 'pacerInterval', 'globalPacingGroup',
            'staticLedsOn', 'dynamicLedsOn', 'staticOnDuration', 'dynamicOnDuration', 'staticKeepAliveScan', 'staticSleepDuration', 'dynamicKeepAliveScan', 'dynamicSleepDuration', 'staticKeepAlivePeriod', 'dynamicKeepAlivePeriod',
            'sub1GhzFrequency', 'sub1GhzOutputPower','otaUpgradeEnabled']
GW_KEYS = ['gwDataSrc', 'gwDataMode']
GW_DATA_SRC = 'gwDataSrc'
GW_DATA_MODE = 'gwDataMode'
GW_KEYS_THIN = ['dataCoupling', 'useStaticLocation', 'gwMgmtMode']
GW_SHARED_KEYS = ['gw_2.4GhzOutputPower', 'gw_txPeriodMs', 'gw_pacerInterval', 'gw_energizingPattern',
                  'gw_rxTxPeriodMs']
TIME_COLUMNS = ['endTimestamp', 'startTimestamp', 'receivedTestConfigTimestamp']
TEST_CONFIG_COLUMNS = ['testId', 'testTimeMins', 'gatewaysIncluded', 'bridgesIncluded']



import random
import pandas as pd

class PacketError:
    def __init__(self):
        pass
            
    @staticmethod
    def _generate_packet_error(duplication):
        # data is sent first
        data_idx = range(0, duplication*2, 2)
        # si is sent after every data packet
        si_idx = range(1, (duplication*2)+1, 2)
        # never drop all data / si packets
        dropped_data = random.sample(data_idx, k=random.randint(0, duplication-1))
        dropped_si = random.sample(si_idx, k=random.randint(0, duplication-1))
        dropped_packets = dropped_data+dropped_si
        drop_packet_by_idx = [True for idx in range(duplication*2)]
        # mark chosen packets as False (dropped)
        for idx in dropped_packets:
            drop_packet_by_idx[idx] = False
        return drop_packet_by_idx


from wiliot_deployment_tools.power_mgmt.power_mgmt import PowerManagementClient
from wiliot_core import check_user_config_is_ok
from wiliot_deployment_tools.common.utils_defines import *
from argparse import ArgumentParser
import sys


def main():
    # parser
    parser = ArgumentParser(prog='wlt-power-mgmt',
                            description='Power Management - CLI Tool for using Wiliot Bridge power management functionality')
    required = parser.add_argument_group('required arguments')
    required.add_argument('-owner', type=str, help="Owner ID", required=True)
    parser.add_argument('-test', action='store_true',
                        help='if flag used, use test environment (prod is used by default)')

    subparsers = parser.add_subparsers()
    parser_enter = subparsers.add_parser('enter',
                                         help='Configure Specified Bridges to work in power management configuration',
                                         epilog="example usage: wlt-power-mgmt -o wiliot enter -brg 0123ABCD -sleepduration 5 -onduration 60")
    parser_enter.set_defaults(func=enter_contex)
    parser_enter.add_argument('-brg', type=str, help="Bridge ID", required=True)
    parser_enter.add_argument('-sleepduration', type=int, help="Sleep duration (minutes)", required=True)
    parser_enter.add_argument('-onduration', type=int, help="On duration (seconds) *rounds to nearest 30 second interval*", required=True)
    parser_enter.add_argument('-keepalive', type=int, help="Keep alive period (seconds) *rounds to nearest 5 second interval* (not required, defaults to 30 seconds)", required=False, default=KEEP_ALIVE_PERIOD)
    parser_enter.add_argument('-scan', type=int, help="Keep alive scan (milliseconds) *rounds to nearest 10 millisecond interval* (not required, defaults to 300 milliseconds)", required=False, default=KEEP_ALIVE_SCAN_DURATION)
    parser_enter.add_argument('-ledoff', action='store_false', help="Configure LEDs off (on by default)", required=False)
    parser_enter.add_argument('-gw', type=str, help="GW ID to configure bridge (required only for broadcast mode)", required=False)
    parser_enter.add_argument('-timeout', type=int, help="Minutes timeout (not required, defaults to 5 minutes)", required=False, default=5)

    parser_exit = subparsers.add_parser(
        'exit', help='Return specified bridges out of power management mode and into normal working mode',
        epilog="example usage: wlt-power-mgmt -o wiliot exit -brg 0123ABCD")
    parser_exit.set_defaults(func=exit_contex)
    parser_exit.add_argument('-brg', type=str, help="Bridge ID", required=True)
    parser_exit.add_argument('-gw', type=str, help="GW ID to configure bridge (not required)", required=False)
    parser_exit.add_argument('-no_config', action='store_false', help="If used, GW will not change to optimal configuration", required=False)
    parser_exit.add_argument('-timeout', type=int, help="Minutes timeout (not required, defaults to 5 minutes)", required=False, default=5)

    parser_mobile = subparsers.add_parser(
        'android', help='Change mobile uplink mode',
        epilog="example usage: wlt-power-mgmt -o wiliot exit -brg 0123ABCD")
    parser_mobile.set_defaults(func=android)
    parser_mobile.add_argument('-gw', type=str, help="Gateway ID", required=True)
    parser_mobile.add_argument('-uplink', choices=['on', 'off'], help='set uplink mode to on or off')

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args, unknown = parser.parse_known_args()
    if args.test:
        env = 'test'
    else:
        env = 'prod'
    owner_id = args.owner

    conf_env = env if env == 'prod' else 'non-prod'
    user_config_file_path, api_key, is_success = check_user_config_is_ok(
        owner_id, conf_env, 'edge')
    if is_success:
        print('credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid credentials - please try again to login')

    args.pm = PowerManagementClient(api_key, owner_id, env)
    args.func(args)


def enter_contex(args):
    res = args.pm.enter_power_mgmt(args.sleepduration, args.onduration, args.keepalive, args.scan, 
                                   leds_on=args.ledoff, gateway_id=args.gw, bridge_id=args.brg, minutes_timeout=args.timeout)
    print(res)


def exit_contex(args):
    res = args.pm.exit_power_mgmt(gateway_id=args.gw, update_gw=args.no_config, 
                                  bridge_id=args.brg, minutes_timeout=args.timeout)
    print(res)

def android(args):
    mode = {'on': True, 'off': False}
    res = args.pm.set_uplink_mode(gateway_id=args.gw, uplink_mode=mode[args.uplink])
    print({args.gw: res})


def main_cli():
    main()
if __name__ == '__main__':
    main()



from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.pkt_generator import BrgPktGeneratorNetwork
from wiliot_deployment_tools.interface.if_defines import *
from wiliot_deployment_tools.interface.packet_error import PacketError
import pkg_resources
import pandas as pd

CSV_NAME = 'packet_table.csv'
PACKET_TABLE_CSV_PATH = pkg_resources.resource_filename(__name__, CSV_NAME)

TEST_COUPLING = 'coupling'
TEST_DOWNLINK = 'downlink'
TEST_UPLINK = 'uplink'

TESTS = [TEST_COUPLING, TEST_UPLINK]
class GeneratedPacketTable:
    
    def __init__(self) -> None:
        self.brg_network = BrgPktGeneratorNetwork()
        self.table = pd.read_csv(PACKET_TABLE_CSV_PATH)
    
    def get_data(self, test, duplication, time_delay, bridge_idx) -> list:    
        assert test in TESTS, 'Invalid Test'
        assert duplication in COUPLING_DUPLICATIONS, 'Invalid Duplication'
        assert time_delay in COUPLING_TIME_DELAYS, 'Invalid Time Delay'
        assert bridge_idx in BRIDGES, 'Invalid Bridge'
        
        t = self.table
        return t.loc[((t['test'] == test) &
                      (t['duplication']==duplication) &
                      (t['time_delay'] == time_delay) &
                      (t['bridge_idx'] == bridge_idx))].to_dict('records')[0]
            
    def _generate_packet_table(self):
        packet_list = []
        
        # COUPLING TEST
        for duplication in COUPLING_DUPLICATIONS:
            debug_print(f'Duplication {duplication}')
            for time_delay in COUPLING_TIME_DELAYS:
                debug_print(f'Time Delay {time_delay}')
                if time_delay == COUPLING_TIME_DELAYS[0]:
                    pkts = self.brg_network.get_new_pkt_pairs()
                    list(map(lambda x: x.update({'time_delay': 0}), pkts))
                else:
                    pkts = self.brg_network.get_scattered_pkt_pairs(time_delay)
                for idx, brg in enumerate(self.brg_network.brg_list):
                    debug_print(f'Bridge {idx}')
                    data = pkts[idx]['data_packet']
                    si = pkts[idx]['si_packet']
                    scattered_time_delay = pkts[idx]['time_delay']
                    packet_error = PacketError()._generate_packet_error(duplication)
                    brg_id = self.brg_network.brg_list[idx].bridge_id
                    # log the sent packet with relevant info from run
                    expected_pkt = brg.get_expected_coupled_mqtt()
                    expected_pkt.update({'duplication': duplication, 'time_delay': time_delay,
                                        'packet_error': packet_error, 'si_rawpacket': si, 'data_rawpacket': data, 'scattered_time_delay': scattered_time_delay})
                    packet_list.append({'test': TEST_COUPLING,
                                        'duplication': duplication,
                                        'time_delay': time_delay,
                                        'bridge_idx': idx,
                                        'packet_error': packet_error,
                                        'expected_mqtt': expected_pkt
                                        ,'data': data, 'si': si, 'bridge_id': brg_id, 'scattered_time_delay': scattered_time_delay
                                        })
        # UPLINK TEST
        for duplication in UPLINK_DUPLICATIONS:
            debug_print(f'Duplication {duplication}')
            for time_delay in UPLINK_TIME_DELAYS:
                debug_print(f'Time Delay {time_delay}')
                pkts = self.brg_network.get_new_pkt_pairs()
                for idx, brg in enumerate(self.brg_network.brg_list):
                    debug_print(f'Bridge {idx}')
                    data = pkts[idx]['data_packet']
                    si = pkts[idx]['si_packet']
                    brg_id = self.brg_network.brg_list[idx].bridge_id
                    # log the sent packet with relevant info from run
                    expected_pkt = brg.get_expected_uncoupled_mqtt()
                    for pkt in expected_pkt:
                        pkt.update({'duplication': duplication, 'time_delay': time_delay})
                    packet_list.append({'test': TEST_UPLINK,
                                        'duplication': duplication,
                                        'time_delay': time_delay,
                                        'bridge_idx': idx,
                                        'expected_mqtt': expected_pkt
                                        ,'data': data, 'si': si, 'bridge_id': brg_id,
                                        })

        pd.DataFrame(packet_list).to_csv(PACKET_TABLE_CSV_PATH)

class CouplingRunData:
    def __init__(self, data) -> None:
        self.test = data['test']
        self.duplication = data['duplication']
        self.time_delay = data['time_delay']
        self.bridge_idx = data['bridge_idx']
        self.packet_error = eval(data['packet_error'])
        self.expected_mqtt = eval(data['expected_mqtt'])
        self.data = data['data']
        self.si = data['si']
        self.bridge_id = data['bridge_id']
        self.scattered_time_delay = data['scattered_time_delay']


    @classmethod
    def get_data(cls, test, duplication, time_delay, bridge_idx):
        packet_data = GeneratedPacketTable().get_data(test, duplication, time_delay, bridge_idx)
        return cls(packet_data)

class UplinkRunData:
    def __init__(self, data) -> None:
        self.test = data['test']
        self.duplication = data['duplication']
        self.time_delay = data['time_delay']
        self.bridge_idx = data['bridge_idx']
        self.expected_mqtt = eval(data['expected_mqtt'])
        self.data = data['data']
        self.si = data['si']
        self.bridge_id = data['bridge_id']   

    @classmethod
    def get_data(cls, test, duplication, time_delay, bridge_idx):
        packet_data = GeneratedPacketTable().get_data(test, duplication, time_delay, bridge_idx)
        return cls(packet_data)

    
if __name__ == "__main__":
    GeneratedPacketTable()._generate_packet_table()



# Files
UT_MQTT_LOG_FILE = "ut_mqtt_log.json"

# GW defines
GW_ID =                         "gatewayId"
ADDITIONAL =                    "additional"
REPORTED_CONF =                 "reportedConf"
GW_CONF =                       "gatewayConf"
GW_NAME =                       "gatewayName"
GW_API_VERSION =                "apiVersion"
LAT =                           "lat"
LNG =                           "lng"


GW_DATA_MODE =                  "gwDataMode"
TAGS_AND_BRGS =                 "Tags & Bridges"
TAGS_ONLY =                     "Tags only"
BRGS_ONLY_38 =                  "Bridges only (ch38)"
BRGS_ONLY_39 =                  "Bridges only (ch39)"
HIBERNATE =                     "Hibernate"

BLE_WIFI =                      "ble_wifi"
BLE_LAN =                       "ble_lan"

GW_ENERGY_PATTERN =             "energizingPattern"
VERSION =                       "version"
WIFI_VERSION =                  "interfaceChipSwVersion"
BLE_VERSION =                   "bleChipSwVersion"
DATA_COUPLING =                 "dataCoupling"
GW_MGMT_MODE =                  "gwMgmtMode"
GW_MODE =                       "gwMode"
ACTIVE =                        "active"
TRANSPARENT =                   "transparent"
PROD =                          "prod"

GET_INFO_ACTION =               "getGwInfo"
REBOOT_GW_ACTION =              "rebootGw"
LOG_PERIOD_ACTION =             "LogPeriodSet"
GET_LOGS =                      "getLogs"
GW_INFO =                       "gatewayInfo"
GW_LOGS =                       "gatewayLogs"
GW_ACTION_ACK =                 "ReceivedAction"
GW_LATITUDE =                   "Latitude"
GW_LONGITUDE =                  "Longitude"
GW_MEL_MODULE_ACK =             "ReceivedMel"
GW_LOG_PERIOD =                 30

# Thin gw defines
THIN_GW_PROTOCOL_VERSION =      "protocolVersion"
THIN_GW_ACTION =                "action"
THIN_GW_GW_ID =                 "gatewayId"
TX_PKT =                        "txPacket"
TX_MAX_DURATION_MS =            "txMaxDurationMs"
TX_MAX_RETRIES =                "txMaxRetries"
TRANPARENT_PKT_LEN =            31 * 2
COUPLED_DATA_PKT_LEN =          29 * 2 # No 1E16

# Bridge defines
BRIDGE_ID =                     "bridgeId"
BRG_STATUS =                    "bridgeStatus"
BRG_STATS =                     "bridgeStats"
BRG_CONFIG =                    "bridgeConfig"
BRG_ACTION =                    "bridgeAction"
BRG_REBOOT =                    "rebootBridge"
BRG_BLINK =                     "blinkBridgeLed"
BRG_ACK_STR =                   GW_ACTION_ACK+"="

OTA_ENABLED =                   "otaUpgradeEnabled"
TX_PROBABILITY =                "txProbability"
TX_REPETITION =                 "txRepetition"
GLOBAL_PACING_GROUP =           "globalPacingGroup"
BOARD_TYPE =                    "boardType"
OUTPUT_POWER_SUB1G =            "sub1GhzOutputPower"
SUB1GHZ_FREQ =                  "sub1GhzFrequency"

FANSTEL =                       "FANSTEL"
MINEW =                         "MINEW"
ENERGOUS =                      "ENERGOUS"
MOKO =                          "MOKO"
UN_INIT =                       "UN_INIT"
DB =                            "DUAL_BAND"
SB =                            "SINGLE_BAND"
BOARD_TYPES_PROFILE_DICT = {0:"FANSTEL_SINGLE_BAND", 1:"FANSTEL_DUAL_BAND", 2:"MINEW_SINGLE_BAND", 3:"MINEW_DUAL_BAND", 4:"ENERGOUS_DUAL_BAND", 0xFFFFFFFF:"UN_INIT"}

# Tags defines
EXT_ID =                        "externalId"
UNKNOWN =                       "unknown"

# Common defines
PACKETS =                       "packets"
TIMESTAMP =                     "timestamp"
ID =                            "id"
CONFIG =                        "config"
ACTION =                        "action"
PAYLOAD =                       "payload"

WLT_SERVER =                    "wiliotServer"
PACER_INTERVAL =                "pacerInterval"
PKT_TYPES_MASK =                "packetTypesMask"
RX_TX_PERIOD =                  "rxTxPeriodMs"
TX_PERIOD =                     "txPeriodMs"
ENERGY_PATTERN =                "energyPattern"
OUTPUT_POWER_2_4 =              "2.4GhzOutputPower"
NFPKT =                         "nfpkt"
RSSI =                          "rssi"
USE_STAT_LOC =                  "useStaticLocation"

# Speedtest
SPEEDTEST_LATENCY =             "SpeedTestLatency"
SPEEDTEST_STRESS =              "SpeedTestStress"
LATENCY_STR =                   "SpeedTestLatency MQTT! Recived"
STRESS_STR =                    "SpeedTestStress MQTT! finished sending"
SPEEDTESTS =                    {SPEEDTEST_LATENCY:LATENCY_STR, SPEEDTEST_STRESS:STRESS_STR}
SPEEDTEST_TIME_LIMIT =          GW_LOG_PERIOD + 20

# External Sensors
IS_SENSOR =                      "isSensor"
IS_EMBEDDED =                    "isEmbedded"
IS_SCRAMBLED =                   "isScrambled"
SENSOR_UUID =                    "sensorServiceId"
SENSOR_ID =                      "sensorId"

ACTION_LONG_TIMEOUT =            120
ACTION_SHORT_TIMEOUT =           5

# Versions
VERSIONS = {
    "1.5.0" : {WIFI_VERSION: "3.5.32", BLE_VERSION: "3.7.25"},
    "1.5.2" : {WIFI_VERSION: "3.5.132", BLE_VERSION: "3.7.25"},
    "1.6.1" : {WIFI_VERSION: "3.5.51", BLE_VERSION: "3.8.18"},
    "1.7.0" : {WIFI_VERSION: "3.9.8", BLE_VERSION: "3.9.24"},
    "1.7.1" : {WIFI_VERSION: "3.10.6", BLE_VERSION: "3.10.13"},
    "1.8.0" : {WIFI_VERSION: "3.11.36", BLE_VERSION: "3.11.40"},
    "1.8.2" : {WIFI_VERSION: "3.11.36", BLE_VERSION: "3.11.42"},
    "1.9.0" : {WIFI_VERSION: "3.12.10", BLE_VERSION: "3.12.36"},
    "1.10.1" : {WIFI_VERSION: "3.13.29", BLE_VERSION: "3.13.25"},
    "3.14.0" : {WIFI_VERSION: "3.14.33", BLE_VERSION: "3.14.64"}
}

# Tests defines
DEFAULT_GW_FIELD_UPDATE_TIMEOUT =   10
DEFAULT_BRG_FIELD_UPDATE_TIMEOUT =  10
HB_PERIOD =                         30
HB_SCAN_TIMEOUT =                   HB_PERIOD + 5
GW_STAT_SCAN_TIMEOUT =              65
BRG_STAT_SCAN_TIMEOUT =             65
BRG_OTA_UPDATE_TIMEOUT =            210
BRG_GLOBAL_PACING_SCAN_TIME =       180
GW_VERSION_UPDATE_DATA_SCAN_TIME =  180
GW_LATITUDE_DEFAULT =               33.0222
GW_LONGITUDE_DEFAULT =              -117.0839
GW_API_VER_DEFAULT =                "200"

# Internal python ut defines - used only in ut
PACER_INTERVAL_THRESHOLD =                          0.90
GLOBAL_PACING_GROUP_THRESHOLD =                     0.70
TX_REPETITION_THRESHOLD =                           0.50
NO_RESPONSE =                                       "NO_RESPONSE"
DONE =                                              "DONE"
TEST_SUCCESS =                                      ":)"
NONE_WILIOT_PKT_CTR =                               "beaconCount"
BASIC_GW_TEST =                                     "basic_gw_test"
MULTI_BRG_TEST =                                    "multibrg_brg_test"
BASIC_BRG_TEST =                                    "basic_brg_test"
ACTION_BRG_TEST =                                   "action_brg_test"
MGMT_PKT =                                          "mgmt_pkt"
SIDE_INFO_PKT =                                     "side_info_pkt"
DECODED_DATA =                                      "decoded_data"
PACKETS_ECHO_OFF =                                  0x10

# Non Default defines
BRG_NON_DEFAULT_RXTX_PERIOD =                       25
BRG_NON_DEFAULT_TX_PERIOD =                         8
BRG_NON_DEFAULT_OUTPUT_POWER_SUB1G =                26
BRG_NON_DEFAULT_PWR_MGMT_KEEP_ALIVE_SCAN =          0
BRG_NON_DEFAULT_TX_REPETITION = 2

LIS2DW12_ACCEL_NON_DEFAULT_STATE_THRESHOLD =        95
LIS2DW12_ACCEL_NON_DEFAULT_WAKE_UP_DURATION =       121
LIS2DW12_ACCEL_NON_DEFAULT_SLEEP_DURATION =         36

# ---------------------------------------------------RTSA defines---------------------------------------------------
# common defines
TRACE_LOG_FILE_NAME =                   "TRACELOG"
TRACE_LOG_FILE_PATH =                   "C:/SignalVu-PC Files/" + TRACE_LOG_FILE_NAME + ".TOV"

# freq defines
FREQ_2_4_GHZ =                          {'37':2.402, '38':2.426, '39':2.480}
FREQ_SUB1G_MHZ =                        {'865_7':865.700, '915':915.000, '916_3':916.300, '917_5':917.500, '918':918.000, '919_1':919.100}

# SignalVu API commands defines
TRACE_DETECTION =                       {'average':'AVERage', 'positive':'POSitive', 'negative':'NEGative', 'positive-negative':'POSNegative', 'sample':'SAMPle'}
MAX_TRACE_POINTS =                      {'1K':'ONEK', '10K':'TENK', '100K':'HUNDredk', 'never_decimate':'NEVerdecimate' }

# default values
DEFAULT_LENGTH_MS =                     30
DEFAULT_TIME_PER_DIVISION_SEC =         5
DEFAULT_RX_TX_PERIOD_SEC =              0.015
BEACON_MIN_LENGTH_SEC =                 375e-6
BEACON_MAX_LENGTH_SEC =                 500e-6
ENERGIZING_TIME_THRESHOLD =             0.3
BEACON_POWER_THRESHOLD =                0.9
BEACON_POWER_CURVE_38 =                 0.7
BEACON_POWER_CURVE_39 =                 0.625
DEFAULT_SPAN_MHZ =                      5

# test times
FREQ_BEACONS_ANALYSIS_TIME_DELTA =      10

# structured energizing patterns information
class energizingPattern:
    def __init__(self, ble_calibration_beacons = [], ble_energy = {}, ble_post_energy_beacons = [], sub1G_energy = False, info = ""):
        self.ble_calibration_beacons = ble_calibration_beacons
        self.ble_energy = ble_energy
        self.ble_post_energy_beacons = ble_post_energy_beacons
        self.sub1G_energy = sub1G_energy
        self.info = info

EP_INFO = {
        '17' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']]),
        '18' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['39'] : 1.0}),
        '20' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['37'] : 0.2, FREQ_2_4_GHZ['39'] : 0.8}),
        '24' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['37'] : 1.0}),
        '25' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['38'] : 1.0}),
        '26' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={2.454 : 1.0}),
        '27' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39'], FREQ_2_4_GHZ['39']]),
        '29' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['38'] : 0.3333, 2.454 : 0.3333, FREQ_2_4_GHZ['39'] : 0.3333}),
        '36' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'],FREQ_2_4_GHZ['38'],FREQ_2_4_GHZ['39']], info="idle"),
        '37' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], 2.415, FREQ_2_4_GHZ['39'], 2.441, 2.428, 2.454, 2.467], ble_energy={2.450 : 1.0}, info="euro"),
        '50' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], sub1G_energy=True),
        '51' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['39'] : 1.0}, sub1G_energy=True),
        '52' : energizingPattern(sub1G_energy=True),
        '55' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['37'] : 1.0}, sub1G_energy=True),
        '56' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['38'] : 1.0}, sub1G_energy=True),
        '57' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={2.454 : 1.0}, sub1G_energy=True),
        '61' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], 2.415, FREQ_2_4_GHZ['39'], 2.441, 2.428, 2.454, 2.467], sub1G_energy=True, info="euro"),
        '62' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], 2.415, FREQ_2_4_GHZ['39'], 2.441, 2.428, 2.454, 2.467], ble_energy={2.475 : 1.0}, sub1G_energy=True, info="euro"),
        '99' : energizingPattern(ble_calibration_beacons=[i/1000.0 for i in range(2402, 2481, 2)])
                }

EP_FREQ_BREAKDOWN_COUNTER_SETUP = {
    2.402 : {'beacons':0, 'energy_in_ms': 0.0},
    2.426 : {'beacons':0, 'energy_in_ms': 0.0},
    2.480 : {'beacons':0, 'energy_in_ms': 0.0},
    2.403 : {'beacons':0, 'energy_in_ms': 0.0},
    2.427 : {'beacons':0, 'energy_in_ms': 0.0},
    2.483 : {'beacons':0, 'energy_in_ms': 0.0},
    2.454 : {'beacons':0, 'energy_in_ms': 0.0},
    2.481 : {'beacons':0, 'energy_in_ms': 0.0},
    2.415 : {'beacons':0, 'energy_in_ms': 0.0},
    2.441 : {'beacons':0, 'energy_in_ms': 0.0},
    2.428 : {'beacons':0, 'energy_in_ms': 0.0},
    2.467 : {'beacons':0, 'energy_in_ms': 0.0},
    2.475 : {'beacons':0, 'energy_in_ms': 0.0},
    0.8657 : {'beacons':0, 'energy_in_ms': 0.0},
    0.915 : {'beacons':0, 'energy_in_ms': 0.0},
    0.9163 : {'beacons':0, 'energy_in_ms': 0.0},
    0.9175 : {'beacons':0, 'energy_in_ms': 0.0},
    0.918 : {'beacons':0, 'energy_in_ms': 0.0},
    0.9191 : {'beacons':0, 'energy_in_ms': 0.0}
}


from wiliot_deployment_tools.interface.if_defines import *
from wiliot_deployment_tools.ag.ut_defines import *

UPLINK_BRG_ID = 'FFFFFFFFFFFF'
RECEIVED = 'received'
SEQ_ID = "sequenceId"
SHARED_COLUMNS = [PAYLOAD]
INT64_COLUMNS = [RSSI]
OBJECT_COLUMNS = [PAYLOAD]
INIT_STAGES_DUPLICATIONS = [i for i in range(2,9)]
REPORT_COLUMNS = ['pkt_id', 'duplication', 'time_delay']
INCREMENTAL_TIME_DELAYS = [10, 50, 100, 255]
TAG_STAGE_ADVA = '0000000000C0'
TAG_STAGE_PACKETS = range(40)
INCREMENTAL_PACKETS = range(255)


from wiliot_deployment_tools.ag.ut_defines import *
from wiliot_deployment_tools.ag.wlt_types_ag import *
class WltPkt():
    supported_pkt_types = WLT_PKT_TYPES

    def __init__(self, hdr=Hdr(), generic=None, pkt=None):
        self.hdr = hdr
        self.generic = generic
        self.pkt = pkt

    def __eq__(self, other):
        if isinstance(other, WltPkt):
            return (
                self.hdr == other.hdr and
                self.generic == other.generic and
                self.pkt == other.pkt
            )
        return False

    def dump(self):
        if self.pkt:
            return self.hdr.dump() + self.pkt.dump()
        return self.hdr.dump() + self.generic.dump()

    def set(self, string):
        if not string.startswith("1E16"):
            string = "1E16" + string

        self.hdr.set(string[0:14])
        if self.hdr.group_id == GROUP_ID_BRG2GW or self.hdr.group_id == GROUP_ID_GW2BRG:
            # GROUP_ID_BRG2GW & GROUP_ID_GW2BRG
            self.generic = GenericV7()
            self.generic.set(string[14:62])

            # MEL modules
            if self.generic.module_type == MODULE_IF and self.generic.msg_type == BRG_MGMT_MSG_TYPE_CFG_INFO:
                self.pkt = ModuleIfV7()
                self.pkt.set(string[14:62])
            elif self.generic.module_type == MODULE_ENERGY_2400:
                self.pkt = ModuleEnergy2400V7()
                self.pkt.set(string[14:62])
            elif self.generic.module_type == MODULE_ENERGY_SUB1G:
                self.pkt = ModuleEnergySub1GV7()
                self.pkt.set(string[14:62])
            elif self.generic.module_type == MODULE_PWR_MGMT:
                self.pkt = ModulePwrMgmtV7()
                self.pkt.set(string[14:62])
            elif self.generic.module_type == MODULE_PERIPH:
                self.pkt = ModulePeriphV7()
                self.pkt.set(string[14:62])
            elif self.generic.module_type == MODULE_CALIBRATION:
                self.pkt = ModuleCalibrationV7()
                self.pkt.set(string[14:62])
            elif self.generic.module_type == MODULE_DATAPATH:
                self.pkt = ModuleDatapathV7()
                self.pkt.set(string[14:62])
            # OLD global config
            elif self.generic.module_type == MODULE_GLOBAL:
                if self.hdr.group_id == GROUP_ID_GW2BRG:
                    # GROUP_ID_GW2BRG
                    if self.generic.msg_type == BRG_MGMT_MSG_TYPE_CFG_SET:
                        self.pkt = Gw2BrgCfgV7()
                        self.pkt.set(string[14:62])
                else:
                    # GROUP_ID_BRG2GW
                    if self.generic.msg_type == BRG_MGMT_MSG_TYPE_CFG_SET or self.generic.msg_type == BRG_MGMT_MSG_TYPE_CFG_INFO:
                        if self.generic.api_version == API_VERSION_V2:
                            self.pkt = Brg2GwCfgV2()
                            self.pkt.set(string[14:62])
                        elif self.generic.api_version == API_VERSION_V5:
                            self.pkt = Brg2GwCfgV5()
                            self.pkt.set(string[14:62])
                        elif self.generic.api_version == API_VERSION_V6:
                            self.pkt = Brg2GwCfgV6()
                            self.pkt.set(string[14:62])
                        else:
                            self.pkt = Brg2GwCfgV7()
                            self.pkt.set(string[14:62])
                    elif self.generic.msg_type == BRG_MGMT_MSG_TYPE_HB:
                        if self.generic.api_version < API_VERSION_V6:
                            self.pkt = Brg2GwHbV1()
                            self.pkt.set(string[14:62])
                        elif self.generic.api_version == API_VERSION_V6:
                            self.pkt = Brg2GwHbV6()
                            self.pkt.set(string[14:62])
                        else:
                            self.pkt = Brg2GwHbV7()
                            self.pkt.set(string[14:62])
                    elif self.generic.msg_type == BRG_MGMT_MSG_TYPE_ACTION:
                        self.pkt = ActionV7()
                        self.pkt.set(string[14:62])
        elif self.hdr.group_id == GROUP_ID_SIDE_INFO_SENSOR:
            self.pkt = SideInfoSensor()
            self.pkt.set(string[14:62])
        elif self.hdr.group_id == GROUP_ID_SIDE_INFO:
            self.pkt = SideInfo()
            self.pkt.set(string[14:62])

hex_str2int = lambda x: int(x,16)


import datetime
from doctest import debug
from struct import pack
import sys
import time
import bitstruct
from wiliot_deployment_tools.api.extended_api import EXCEPTIONS_TO_CATCH, AndroidGatewayAction, BridgeThroughGatewayAction, ExtendedEdgeClient, GatewayType
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.common.utils_defines import KEEP_ALIVE_PERIOD, KEEP_ALIVE_SCAN_DURATION, SEC_TO_SEND, BROADCAST_DST_MAC, EXIT_POWER_MGMT_GW_DICT


class PowerManagementError(Exception):
    pass


class PowerManagementPacket:
    def __init__(self, payload=None):
        """
        :type payload: str
        :param payload: power management packet payload (hex string)
        """
        self.payload = payload

    def __repr__(self):
        sleep_duration_mins, on_duration_secs, keep_alive_period_secs, keep_alive_scan_ms, leds_on = self.unpack()
        return f'SD[{sleep_duration_mins}]OD[{on_duration_secs}]KA_P[{keep_alive_period_secs}]KA_S[{keep_alive_scan_ms}]L[{leds_on}]'

    @classmethod
    def from_args(cls, sleep_duration_mins, on_duration_secs,
                  keep_alive_period_secs=KEEP_ALIVE_PERIOD, keep_alive_scan_ms=KEEP_ALIVE_SCAN_DURATION, leds_on=True,
                  silent=False):
        """
        Create power management packet from specified arguments
        :type sleep_duration_mins: int
        :param sleep_duration_mins: Sleep duration (minutes)
        :type on_duration_secs: int
        :param on_duration_secs: On duration (seconds) - rounds to the nearest 30 second interval
        :type keep_alive_period_secs: int
        :param keep_alive_period_secs: Keep-alive period (seconds) - rounds to the nearest 5 second interval
        :type keep_alive_scan_ms: int
        :param keep_alive_scan_ms: Keep-alive scan duration (milliseconds) - rounds to the nearest 10 millisecond interval
        :type leds_on: bool
        :param leds_on: leds on/off during power management
        :type silent: bool
        :param silent: print power management
        :rtype: PowerManagementPacket
        :return: packet of specified parameters
        """
        if keep_alive_period_secs is None:
            keep_alive_period_secs = KEEP_ALIVE_PERIOD
        if keep_alive_scan_ms is None:
            keep_alive_scan_ms = KEEP_ALIVE_SCAN_DURATION
        payload = cls.pack_pwr_mgmt_struct(sleep_duration_mins, on_duration_secs, keep_alive_period_secs,
                                           keep_alive_scan_ms, leds_on, silent)
        packet = cls(payload)
        return packet

    @classmethod
    def exit_packet(cls, silent=True):
        """
        returns packet with 0 values (to exit power management mode)
        :type silent: bool
        :param silent: print power management
        :rtype: PowerManagementPacket
        :return: packet to exit power management
        """
        return cls.from_args(sleep_duration_mins=0,
                             on_duration_secs=0,
                             keep_alive_period_secs=0,
                             keep_alive_scan_ms=0,
                             leds_on=True,
                             silent=silent)

    def unpack(self):
        return self.unpack_pwr_mgmt_packet(self.payload)

    @staticmethod
    def pack_pwr_mgmt_struct(sleep_duration_mins, on_duration_secs,
                             keep_alive_period_secs, keep_alive_scan_ms, leds_on, silent=False):
        """
        Create power management packet from specified arguments
        :type sleep_duration_mins: int
        :param sleep_duration_mins: Sleep duration (minutes)
        :type on_duration_secs: int
        :param on_duration_secs: On duration (seconds) - rounds to the nearest 30 second interval
        :type keep_alive_period_secs: int
        :param keep_alive_period_secs: Keep-alive period (seconds) - rounds to the nearest 5 second interval
        :type keep_alive_scan_ms: int
        :param keep_alive_scan_ms: Keep-alive scan duration (milliseconds) - rounds to the nearest 10 millisecond interval
        :type leds_on: bool
        :param leds_on: leds on/off during power management
        :type silent: bool
        :param silent: print power management packet parameters
        :rtype: str
        :return: hex string of packet
        """
        if sleep_duration_mins != 0 and on_duration_secs != 0:
            assert keep_alive_period_secs < sleep_duration_mins * 60, 'Keep alive period must be shorter than sleep period!'
        if not 0 <= sleep_duration_mins <= 2047:
            raise PowerManagementError('Sleep Duration (minutes) must be a value between 0 and 2047!')
        if not 0 <= on_duration_secs <= 3810:
            raise PowerManagementError('On Duration (seconds) must be a value between 0 and 3810!')
        if not 0 <= keep_alive_period_secs <= 155:
            raise PowerManagementError('Keep Alive Period (seconds) must be a value between 0 and 155!')
        if not 0 <= keep_alive_scan_ms <= 630:
            raise PowerManagementError('Keep Alive Scan (milliseconds) must be a value between 0 and 630!')

        sleep_duration = sleep_duration_mins
        on_duration = int(on_duration_secs / 30)
        keep_alive_period = int(keep_alive_period_secs / 5)
        keep_alive_scan = int(keep_alive_scan_ms / 10)
        if sleep_duration != 0 and on_duration == 0 and keep_alive_period == 0:
            raise PowerManagementError('Please set keep alive period!')

        if not silent:
            debug_print(f'''Configuring Power Management:
                  Sleep Duration: {sleep_duration} minutes
                  On Duration: {on_duration * 30} seconds
                  Keep Alive Period: {keep_alive_period * 5} seconds
                  Keep Alive Scan: {keep_alive_scan * 10} milliseconds
                  LEDs: {'ON' if leds_on else 'OFF'}
                  ''')

        fmt = 'p2 b1 u6 u5 u7 u11'
        assert 0 <= keep_alive_scan <= 63
        assert 0 <= keep_alive_period <= 31
        assert 0 <= on_duration <= 127
        assert 0 <= sleep_duration <= 2047
        payload = bitstruct.pack(fmt,
                                 leds_on,
                                 keep_alive_scan,
                                 keep_alive_period,
                                 on_duration,
                                 sleep_duration).hex().upper()

        return payload

    @staticmethod
    def unpack_pwr_mgmt_packet(payload):
        """
        unpack packet to arguments
        :type payload: str
        :param payload: hex string of packet
        :rtype: tuple
        :return: (sleep_duration_mins, on_duration_secs, keep_alive_period_secs, keep_alive_scan_ms, leds_on)
        """
        assert len(payload) >= 8
        payload = payload[:8]
        payload = bytes.fromhex(payload)
        fmt = 'p2 b1 u6 u5 u7 u11'
        leds_on, keep_alive_scan, keep_alive_period, on_duration, sleep_duration = \
            bitstruct.unpack(fmt, payload)
        sleep_duration_mins = sleep_duration
        on_duration_secs = on_duration * 30
        keep_alive_period_secs = keep_alive_period * 5
        keep_alive_scan_ms = keep_alive_scan * 10
        return sleep_duration_mins, on_duration_secs, keep_alive_period_secs, keep_alive_scan_ms, leds_on


class BridgePowerManagementClient(ExtendedEdgeClient):

    
    def get_relevant_bridges(self, gateway_id):
        """
        Returns seen bridges by gateway that are associated to owner and have ble version with power management support
        :type gateway_id: str
        :param gateway_id: gateway ID
        :rtype: list
        :return: list of relevant bridge IDs
        """
        bridges = self.get_seen_bridges(gateway_id)
        relevant_bridges = bridges
        for bridge in bridges:
            brg_ble = self.get_brg_ble_version(bridge)
            if not self.check_ble_power_mgmt_support(brg_ble):
                # debug_print(f'Bridge {bridge} does not support power management!')
                relevant_bridges.remove(bridge)
        return relevant_bridges


    @staticmethod
    def check_ble_power_mgmt_support(ble_ver):
        """
        checks if power management is supported for BLE version
        :rtype: bool
        :return: True if supported, False otherwise
        """
        ble_ver = ble_ver.split('.')
        if not ((int(ble_ver[0]) > 3) or
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 11) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 10 and int(ble_ver[2]) > 39)):
            return False
        else:
            return True


    def send_pwr_mgmt_pkt(self, bridge_id, packet, gateway_id=None,
                          silent=False, seconds_to_send=1):
        """
        send power management packet to bridge
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: PowerManagementPacket
        :param packet: packet to send to bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type silent: bool
        :param silent: print power management packet parameters
        :type seconds_to_send: int
        :param seconds_to_send: seconds to repeat packet
        :rtype: bool
        :return: True if sent, False otherwise
        """
        if gateway_id is None:
            via_gw = self.get_bridge_relevant_gw(bridge_id)
        else:
            via_gw = gateway_id
        seconds_to_send = datetime.timedelta(seconds=seconds_to_send)
        # Check BRG Compatability
        brg_version = None
        if bridge_id is not BROADCAST_DST_MAC:
            brg_version = self.get_brg_ble_version(bridge_id)
        # Check GW Compatability
        gw_type = self.get_gateway_type(via_gw)
        if gw_type == GatewayType.WIFI:
            gw_version = self.get_gw_ble_version(via_gw)
            for version in filter(None, [brg_version, gw_version]):
                if not self.check_ble_power_mgmt_support(version):
                    return False
            rxtx = datetime.timedelta(milliseconds=
                                    self.get_gateway(via_gw)['reportedConf']['additional']['rxTxPeriodMs'])
        if gw_type == GatewayType.MOBILE:
            rxtx = datetime.timedelta(milliseconds=140)
        reps = seconds_to_send // rxtx + 1
        if not silent:
            debug_print(f'Sending packet through GW {via_gw} for {seconds_to_send.total_seconds()} seconds, rxTx = {int(rxtx.total_seconds()*1000)} ms, reps = {reps}')
        try:
            res = self.send_bridge_action_through_gw(gateway_id,
                                                 BridgeThroughGatewayAction.POWER_MGMT,
                                                 packet.payload, bridge_id, reps=reps)
        except EXCEPTIONS_TO_CATCH as e:
            debug_print('Exception caught when sending power mgmt packet!')
            debug_print(e)
            return False
        return res

    def broadcast_pwr_mgmt_packet(self, packet, gateway_id, silent=False, seconds_to_send=1):
        """
        broadcast packet to all bridges connected to GW
        :type packet: PowerManagementPacket
        :param packet: packet to broadcast
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type silent: bool
        :param silent: print power management packet parameters
        :rtype: tuple
        :return: results (bool), bridges sent to (list)
        """
        if not self.check_gw_compatible_for_action(gateway_id):
            raise PowerManagementError(f'Gateway {gateway_id} not compatible with power management! '
                                       f'Try to update GW version')
        brgs_to_broadcast = self.get_relevant_bridges(gateway_id)
        if not silent:
            debug_print(f'Broadcasting Packet to bridges {brgs_to_broadcast}')
        res = self.send_pwr_mgmt_pkt(BROADCAST_DST_MAC, packet, gateway_id, silent, seconds_to_send)
        return res, brgs_to_broadcast

    def check_pwr_mgmt_received(self, gateway_id, bridge_id, packet, time_sent):
        """
        check if power management packet was received by bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: PowerManagementPacket
        :param packet: packet to check
        :type time_sent: datetime
        :param time_sent: time packet was sent
        :rtype bool
        :return: True if packet was received, None otherwise
        """
        hours_back = (datetime.datetime.now() - time_sent + datetime.timedelta(seconds=5)).total_seconds() / 3600
        acks = self.get_acks(gateway_id, hours_back)
        if acks is None:
            return None
        acks = acks[(acks['actionType'] == BridgeThroughGatewayAction.POWER_MGMT) &
                    (acks['bridgeId'] == bridge_id) &
                    (acks['payload'].apply(lambda x: PowerManagementPacket(x[:8]).payload)
                     == packet.payload)]
        if len(acks) > 0:
            return acks.to_dict()
        return None

    def check_pwr_mgmt_broadcast_received(self, gateway_id, bridges, packet, time_sent):
        """
        check if power management broadcast was received by bridges
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridges: list
        :param bridges: list of Bridge IDs
        :type packet: PowerManagementPacket
        :param packet: packet to check
        :type time_sent: datetime
        :param time_sent: time packet was sent
        :rtype: dict
        :return: dict of bridgeIDs:received status
        """
        hours_back = (datetime.datetime.now() - time_sent + datetime.timedelta(seconds=10)).total_seconds() / 3600
        acks = self.get_acks(gateway_id, hours_back)
        result = dict().fromkeys(bridges)
        for brg in bridges:
            if acks is None:
                result[brg] = False
                continue
            brg_acks = acks[(acks['actionType'] == BridgeThroughGatewayAction.POWER_MGMT) &
                        (acks['bridgeId'] == brg) &
                        (acks['payload'].apply(lambda x: PowerManagementPacket(x[:8]).payload)
                         == packet.payload)]
            if len(brg_acks) > 0:
                result[brg] = True
            else:
                result[brg] = False
        return result

    def send_packet_until_ack(self, gateway_id, bridge_id, packet, minutes_timeout=3):
        """
        send power management packet repeatedly until it is received by the bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: PowerManagementPacket
        :param packet: packet to send
        :type minutes_timeout: int
        :param minutes_timeout: timeout
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        time_sent = datetime.datetime.now()
        debug_print(f'Configuring {bridge_id} To {packet} | Max Waiting time {minutes_timeout} minutes')
        self.send_pwr_mgmt_pkt(bridge_id, packet, gateway_id, silent=False, seconds_to_send=SEC_TO_SEND)
        time_gw_stops = time_sent + datetime.timedelta(seconds=SEC_TO_SEND)
        time_end = time_sent + datetime.timedelta(minutes=minutes_timeout)
        curr_time = time_sent
        ack = self.check_pwr_mgmt_received(gateway_id, bridge_id, packet, time_sent)
        print('Waiting For Ack...', end='', flush=True)
        while curr_time < time_end and ack is None:
            curr_time = datetime.datetime.now()
            if curr_time > time_gw_stops - datetime.timedelta(seconds=1):  # 1 second buffer
                self.send_pwr_mgmt_pkt(bridge_id, packet, gateway_id, silent=True, seconds_to_send=SEC_TO_SEND)
                time_gw_stops = datetime.datetime.now() + datetime.timedelta(seconds=SEC_TO_SEND)
                print('!', end='')
                sys.stdout.flush()
            time.sleep(0.5)
            print('.', end='')
            sys.stdout.flush()
            ack = self.check_pwr_mgmt_received(gateway_id, bridge_id, packet, time_sent)
        print()  # enter newline
        ack_received = False
        if ack is not None:
            ack_received = True
            debug_print(f'Ack Received after {datetime.datetime.now() - time_sent}')
        else:
            debug_print(f'Ack not received after {datetime.datetime.now() - time_sent}!')
        return {bridge_id: ack_received}

    def broadcast_packet_until_ack(self, gateway_id, packet, minutes_timeout=3):
        """
        send power management packet repeatedly until it is received by the all 'seen' bridges
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type packet: PowerManagementPacket
        :param packet: packet to send
        :type minutes_timeout: int
        :param minutes_timeout: timeout
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        time_sent = datetime.datetime.now()
        res, bridges = self.broadcast_pwr_mgmt_packet(packet, gateway_id, silent=False, seconds_to_send=SEC_TO_SEND)
        debug_print(f'Configuring {bridges} To {packet} | Max Waiting time {minutes_timeout} minutes')
        time_gw_stops = time_sent + datetime.timedelta(seconds=SEC_TO_SEND)
        time_end = time_sent + datetime.timedelta(minutes=minutes_timeout)
        curr_time = time_sent
        ack = self.check_pwr_mgmt_broadcast_received(gateway_id, bridges, packet, time_sent)
        print('Waiting For Ack...', end='', flush=True)
        while curr_time < time_end and not all(ack.values()):
            curr_time = datetime.datetime.now()
            if curr_time > (time_gw_stops - datetime.timedelta(seconds=1)):  # 1 second buffer
                new_bridges = self.get_relevant_bridges(gateway_id)
                if bridges != new_bridges:
                    bridges = new_bridges
                    debug_print(f'Bridges seen by GW changed while broadcasting! now broadcasting to {bridges}')
                self.broadcast_pwr_mgmt_packet(packet, gateway_id, silent=True, seconds_to_send=SEC_TO_SEND)
                time_gw_stops = datetime.datetime.now() + datetime.timedelta(seconds=SEC_TO_SEND)
                print('!', end='')
                sys.stdout.flush()
            time.sleep(0.5)
            print('.', end='')
            sys.stdout.flush()
            ack = self.check_pwr_mgmt_broadcast_received(gateway_id, bridges, packet, time_sent)
        print()  # enter newline
        if not all(ack.values()):
            # list of bridges which did not return ack
            brgs_missed = list(dict(filter(lambda x: x[1] is False, ack.items())).keys())
            debug_print(f'Packet not received by bridges {brgs_missed} after {minutes_timeout} minutes!')
        else:
            debug_print(
                f'Ack Received after {datetime.datetime.now() - time_sent} by bridges {list(ack.keys())}')
        return ack

    def enter_power_mgmt(self, sleep_duration_mins, on_duration_secs, keep_alive_period_secs=None,
                         keep_alive_scan_ms=None, leds_on=True, gateway_id=None, bridge_id=None, broadcast=False,
                         minutes_timeout=3):
        """
        Enter power management mode with the specified parameters.
        Function will search for the relevant gateway ID (if not specified) to send the configuration packet through.
        :type sleep_duration_mins: int
        :param sleep_duration_mins: Sleep duration (minutes)
        :type on_duration_secs: int
        :param on_duration_secs: On duration (seconds) - rounds to the nearest 30 second interval
        :type keep_alive_period_secs: int
        :param keep_alive_period_secs: Keep-alive period (seconds) - rounds to the nearest 5 second interval.
        Defaults to 30 seconds
        :type keep_alive_scan_ms: int
        :param keep_alive_scan_ms: Keep-alive scan duration (milliseconds) - rounds to the nearest 10 millisecond interval.
        Defaults to 150 milliseconds
        :type leds_on: bool
        :param leds_on: leds on/off during power management
        :type gateway_id: str
        :param gateway_id: Gateway ID (optional)
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type broadcast: bool
        :param broadcast: if True, broadcast power mgmt to all 'seen' bridges
        :type minutes_timeout: int
        :param minutes_timeout: minutes timeout
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        packet = PowerManagementPacket.from_args(sleep_duration_mins=sleep_duration_mins,
                                                 on_duration_secs=on_duration_secs,
                                                 keep_alive_period_secs=keep_alive_period_secs,
                                                 keep_alive_scan_ms=keep_alive_scan_ms,
                                                 leds_on=leds_on, silent=False)
        if not ((bridge_id is not None and broadcast is False) or (bridge_id is None and broadcast is True)):
            raise PowerManagementError('Must specify bridgeId / set broadcast to True')
        if broadcast:
            if gateway_id is None:
                raise PowerManagementError('Must specify GW ID to broadcast')
            return self.broadcast_packet_until_ack(gateway_id, packet, minutes_timeout)
        via_gw = self.get_bridge_relevant_gw(bridge_id) if gateway_id is None else gateway_id
        if not self.check_gw_compatible_for_action(via_gw):
            raise PowerManagementError(f'GW {via_gw} Not compatible for power management!')
        res = self.send_packet_until_ack(via_gw, bridge_id, packet, minutes_timeout)
        if False in res.values():
            ###HOTFIX
            if self.get_gateway_type(via_gw) == GatewayType.MOBILE:
                self.send_packet_until_ack(via_gw, bridge_id, PowerManagementPacket.exit_packet(), minutes_timeout)
                res = self.send_packet_until_ack(via_gw, bridge_id, packet, minutes_timeout)
                if False in res.values():
                    raise PowerManagementError(f'Failed to exit power management! {res}')
                else:
                    return res
            else:
                raise PowerManagementError(f'Failed to exit power management! {res}')
        return res

    def exit_power_mgmt(self, gateway_id=None, update_gw=False, bridge_id=None, broadcast=False, minutes_timeout=3):
        """
        Exit power management mode (to normal bridge behavior).
        Function will search for the relevant gateway ID (if not specified) to send the configuration packet through.
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type gateway_id: str
        :param gateway_id: Gateway ID (optional)
        :type update_gw: bool
        :param update_gw: Update GW to optimal configuration to send configuration packet
        :type minutes_timeout: int
        :param minutes_timeout: minutes timeout
        :type broadcast: bool
        :param broadcast: if True, broadcast power mgmt to all 'seen' bridges
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        packet = PowerManagementPacket().exit_packet(silent=False)
        if not ((bridge_id is not None and broadcast is False) or (bridge_id is None and broadcast is True)):
            raise PowerManagementError('Must specify bridgeId / set broadcast to True')
        if broadcast:
            if gateway_id is None:
                raise PowerManagementError('Must specify GW ID to broadcast')
        via_gw = self.get_bridge_relevant_gw(bridge_id) if gateway_id is None else gateway_id
        if not self.check_gw_compatible_for_action(via_gw):
            raise PowerManagementError(f'GW {via_gw} Not compatible for power management!')
        gateway_type = self.get_gateway_type(via_gw)
        need_to_update = False
        if gateway_type == GatewayType.WIFI:
            gw_dict = self.get_gateway(via_gw)['reportedConf']['additional']
            for param in EXIT_POWER_MGMT_GW_DICT:
                if gw_dict[param] != EXIT_POWER_MGMT_GW_DICT[param]:
                    need_to_update = True
            if need_to_update and update_gw:
                debug_print('Changing GW Params to rxTxPeriod = 15 ms, gwRxChannel = Ch. 39')
            self.change_gw_config([via_gw], EXIT_POWER_MGMT_GW_DICT)
        if broadcast:
            res = self.broadcast_packet_until_ack(gateway_id, packet, minutes_timeout)
        else:
            res = self.send_packet_until_ack(via_gw, bridge_id, packet, minutes_timeout)
        if need_to_update and update_gw:
            debug_print('Finished sending exit packets, returning GW to old state')
            self.change_gw_config([via_gw], gw_dict)
        if False in res.values():
                raise PowerManagementError(f'Failed to exit power management! {res}')
        return res            


class AndroidPowerManagementClient(ExtendedEdgeClient):
    def set_uplink_mode(self, gateway_id, uplink_mode):
        """
        set uplink mode (on/off) for android GW
        :param gateway_id: gateway ID
        :type gateway_id: str
        :param uplink_mode: uplink mode
        :type uplink_mode: bool
        """
        payload = {True: AndroidGatewayAction.ENABLE_UPLINK,
                   False: AndroidGatewayAction.DISABLE_UPLINK}
        gw_type = self.get_gateway_type(gateway_id)
        assert gw_type == GatewayType.MOBILE, f'GW {gateway_id} not an android GW!'
        def send_uplink_mode_and_sleep(gateway_id, uplink_mode):
            debug_print(f'Setting GW {gateway_id} uplink to {uplink_mode}...')
            res = self.send_action_to_gateway(gateway_id, payload[uplink_mode])
            debug_print('Waiting for GW to update... (30 seconds)')
            time.sleep(30)
            return res
        tries = 1
        if uplink_mode:
            tries = 3
        for num in range(tries):
            res = send_uplink_mode_and_sleep(gateway_id, uplink_mode)
        return res

class PowerManagementClient(BridgePowerManagementClient, AndroidPowerManagementClient):
    pass





from wiliot_deployment_tools.connectivity_analyzer.connectivity_analyzer import GwHbActionClient
from wiliot_core import check_user_config_is_ok
from wiliot_deployment_tools.common.utils_defines import *
from argparse import ArgumentParser
import sys
import bitstruct


def main():
    # parser
    parser = ArgumentParser(prog='wlt-connectivity-analyzer',
                            usage='%(prog)s -o OwnerID -brg BridgeID (-t) [mode]',
                            description='CLI Tool for using Wiliot Bridge calibration management functionality')
    parser.add_argument('-owner', type=str, help="Owner ID", required=True)
    parser.add_argument('-t', action='store_true',
                        help='if flag used, use test environment (prod is used by default)')
    parser.add_argument('-gws', nargs='+', type=str, help="gateways id list to check connectivity", required=False)
    parser.add_argument('-brgs', nargs='+', type=str, help="bridges id list to check connectivity", required=False)
    parser.add_argument('-location', type=str, help="Location/Zone to search bridges", required=False)
    parser.add_argument('-bridge_log', action='store_true', help="print all rssi from all seen gw for a given bridge", required=False)
    parser.add_argument('-m', type=int, help="Minutes timeout (not required, defaults to 5 minutes)", required=False, default=5)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args, unknown = parser.parse_known_args()
    if not args.location and not args.gws and not args.brgs:
        print('Error: must insert location, bridges list or gateways list')
        return False
    if (args.location and args.gws) or (args.location and args.brgs) or (args.brgs and args.gws):
        print('Error: too many arguments, Choose: Location, bridges or gateways')
        return False
    if args.t:
        env = 'test'
    else:
        env = 'prod'
    owner_id = args.owner
    conf_env = env if env == 'prod' else 'non-prod'

    if args.bridge_log:
        bridge_log = True
    else:
        bridge_log = False

    user_config_file_path, asset_api_key, is_success = check_user_config_is_ok(owner_id, conf_env, 'asset')
    if is_success:
        print('Asset Management credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid Asset Management credentials - please try again to login')
    user_config_file_path, edge_api_key, is_success = check_user_config_is_ok(owner_id, conf_env, 'edge')
    if is_success:
        print('Edge Management credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid Edge Management credentials - please try again to login')

    
    client = GwHbActionClient(edge_api_key,asset_api_key, owner_id, env)
    client.bridge_gw_downlink_analyzer(location=args.location, minutes_timeout=args.m,
                                       bridges_list=args.brgs,gws_list = args.gws,is_bridge_log = bridge_log)


def main_cli():
    main()
if __name__ == '__main__':
    main()



import os
from os import getenv
from sys import argv
import boto3
import shutil
import re

BUCKET_NAME = "wiliot-partner-pkg"


VERSION = ".".join(os.getenv("version").split(".")[:-1]).split("+")[0]


def delete_argv_folders():
    folders_to_delete = ['internal', 'wiliot.egg-info']
    for folder in os.getenv('DELETED_FOLDERS').split(','):
        folders_to_delete.append(folder)

    files = [f for f in os.listdir('.') if os.path.isfile(f)]
    for f in files:
        if f == "setup.py" or "ci.py":
            pass
        else:
            os.remove(f)
            print(f"Deleting the '{f}' file")

    for root, subdirs, files in os.walk("wiliot/"):
        for folder in folders_to_delete:
            if re.search(f".*{folder}.*", root):
                print(f"Deleting the '{root}' folder")
                shutil.rmtree(root)


def upload_latest_version(filepath: str, filename: str, partner_name: str):
    s3.meta.client.upload_file(filepath, 'wiliot-partner-pkg', f'{partner_name}/latest/{filename}')
    print("Uploaded: ", filepath, 'wiliot-partner-pkg', f'{partner_name}/latest/{filename}')
    s3.meta.client.upload_file(filepath, 'wiliot-partner-pkg', f'{partner_name}/previous_releases/{filename}')
    print("Uploaded: ", filepath, 'wiliot-partner-pkg', f'{partner_name}/previous_releases/{filename}')


def earth_latest_folder(partner_name: str):
    bucket = s3.Bucket(BUCKET_NAME)
    for obj in bucket.objects.filter(Prefix=f'{partner_name}/latest/'):
        old_filename = obj.key.split("/")[2]
        print(s3.meta.client.delete_object(Bucket=BUCKET_NAME, Key=f'{partner_name}/latest/{old_filename}'))


def release_version(partner_name: str):
    filename = f'wiliot-{".".join(os.getenv("version").split("."))}.tar.gz'
    filename_in_bucket = f'wiliot-{VERSION}+{os.getenv("BITBUCKET_BUILD_NUMBER")}.tar.gz'
    print(os.getenv("version"))
    print(filename, partner_name, filename_in_bucket)
    earth_latest_folder(partner_name)
    upload_latest_version(f'dist/{filename}', filename_in_bucket, partner_name)


def safe_list_get(li, index, default_value=None):
    try:
        return li[index]
    except IndexError:
        return default_value


if __name__ == '__main__':
    if safe_list_get(argv, 1) == "DELETED_FOLDERS":
        delete_argv_folders()
    else:
        s3 = boto3.resource("s3",
                            aws_access_key_id=getenv("AWS_ACCESS_KEY_ID"),
                            aws_secret_access_key=getenv("AWS_SECRET_ACCESS_KEY"),
                            aws_session_token=getenv("AWS_SESSION_TOKEN"))
        release_version(os.getenv("PARTNER_NAME"))



import datetime
import os
from wiliot_deployment_tools.api.extended_api import ExtendedEdgeClient
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.if_defines import SEP
from wiliot_deployment_tools.interface.mqtt import MqttClient

PASS_STATUS = {True: 'PASS', False: 'FAIL'}

class GenericTest:
    def __init__(self, mqttc: MqttClient, ble_sim: BLESimulator, 
                 gw_capabilities:GWCapabilities, gw_id, owner_id, test_name, **kwargs):
        # Clients
        self.mqttc = mqttc
        self.ble_sim = ble_sim
        
        # Test-Related
        self.gw_capabilities = gw_capabilities
        self.report = ''
        self.report_html = ''
        self.test_pass = False
        self.start_time = None
        self.test_name = test_name
        self.test_dir = os.path.join(self.certificate_dir, self.test_name)
        self.env_dirs.create_dir(self.test_dir)
        self.stages = []
        
    def __repr__(self):
        return self.test_name
    
    def run(self):
        self.start_time = datetime.datetime.now()
        debug_print(f"Starting Test {self.test_name} : {self.start_time}")
        
    def runtime(self):
        return datetime.datetime.now() - self.start_time
    
    def add_to_test_report(self, report):
        self.report += '\n' + report
    
    def create_test_html(self):
        self.report_html = self.template_engine.render_template('test.html', 
                                                                test_name=self.test_name,
                                                                test_pass = self.test_pass,
                                                                running_time = self.runtime(),
                                                                stages=[stage.report_html for stage in self.stages])
    
    def end_test(self):
        # TODO - remove
        # debug_print(f'\n{SEP}')
        # debug_print(f'{self.test_name} {PASS_STATUS[self.test_pass]}, Running time {self.runtime()}')
        # debug_print(self.report)
        self.create_test_html()
class GenericStage():
    def __init__(self, stage_name, **kwargs):
        #Stage Params
        self.stage_name = stage_name
        self.stage_pass = False
        self.report = ''
        self.report_html = ''
        self.start_time = None
        self.csv_path = os.path.join(self.test_dir, f'{self.stage_name}.csv')
        
    def __repr__(self):
        return self.stage_name
    
    def prepare_stage(self):
        debug_print(f'### Starting Stage: {self.stage_name}')

    def add_to_stage_report(self, report):
        self.report += f'{report}\n'
    
    def generate_stage_report(self):
        return self.report



import pandas as pd
from wiliot_deployment_tools.interface.uart_if import UARTInterface
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.if_defines import *
import time
import datetime
import threading
import binascii

class SnifferPkt():
    def __init__(self, raw_output, time_received, rx_channel):
        self.adva = raw_output[:12]
        self.packet = raw_output[12:74]
        self.rssi = int.from_bytes(binascii.unhexlify(raw_output[74:]), 'big') * -1
        self.time_received = time_received
        self.rx_channel = rx_channel
    
    def __repr__(self):
        return f'CH{self.rx_channel}|{self.adva}{self.packet} RSSI:{self.rssi} {self.time_received}'
    
    def to_dict(self):
        return {'adva': self.adva, 'packet': self.packet, 'rssi': self.rssi, 'time_received': self.time_received, 'rx_channel': self.rx_channel}

class SnifferPkts():
    def __init__(self, pkts=[]):
        self.pkts = pkts
    
    def __add__(self, other):
        return SnifferPkts(self.pkts + other.pkts)
        
    def __len__(self):
        return len(self.pkts)
    
    def __repr__(self):
        return self.pkts
    
    def process_pkt(self, raw_output, time_received, rx_channel, print_pkt=False):
        pkt = SnifferPkt(raw_output, time_received, rx_channel)
        self.pkts.append(pkt)
        if print_pkt:
            print(pkt)
        
    def filter_pkts(self, raw_packet=None, adva=None, time_range:tuple=None):
        result = []
        for pkt in self.pkts:
            if (raw_packet is not None) and (pkt.packet == raw_packet) or \
                (adva is not None) and (pkt.adva == adva) or \
                (time_range is not None) and (time_range[0] < pkt.time_received < time_range[1]):
                result.append(pkt)
        return SnifferPkts(result)
    
    def flush_pkts(self):
        self.pkts = []
        
    def to_list(self):
        return [p.to_dict() for p in self.pkts]
    
    def to_pandas(self):
        return pd.DataFrame().from_dict(self.to_list())

class BLESniffer():
    def __init__(self, uart:UARTInterface, print_pkt=False):
        self.uart = uart
        self.listener_thread = None
        self.sniffer_pkts = SnifferPkts()
        self.listening = False
        self.listener_lock = threading.Lock()
        self.print = print_pkt
        self.rx_channel = 0

    def packet_listener(self):
        while self.listening:
            line = self.uart.read_line()
            if line is not None and len(line) == 76:
                with self.listener_lock:
                    self.sniffer_pkts.process_pkt(line, datetime.datetime.now(), self.rx_channel, self.print)
    
    # Change sniffing modes
    def start_sniffer(self, rx_channel):
        self.uart.set_sniffer(rx_channel)
        self.rx_channel = rx_channel
        self.listener_thread = threading.Thread(target=self.packet_listener)
        self.listening = True
        self.listener_thread.start()
        
    def stop_sniffer(self):
        self.flush_pkts()
        self.listening = False
        if self.listener_thread is not None:
            self.listener_thread.join()
            self.listener_thread = None
            self.uart.cancel_sniffer()
    
    def reset_sniffer(self, rx_channel):
        self.stop_sniffer()
        self.start_sniffer(rx_channel)
        self.flush_pkts()
        
    # Data Handling
    def get_all_pkts(self):
        return self.sniffer_pkts
    
    def get_filtered_packets(self, raw_packet=None, adva=None, time_range:tuple=None):
        return self.sniffer_pkts.filter_pkts(raw_packet, adva, time_range)
    
    def flush_pkts(self):
        self.sniffer_pkts.flush_pkts()
        
    def to_pandas(self):
        return self.sniffer_pkts.to_pandas()

class BLESnifferContext():
    def __init__(self, ble_sniffer:BLESniffer, rx_channel):
        self.ble_sniffer = ble_sniffer
        self.rx_channel = rx_channel

    def __enter__(self):
        self.ble_sniffer.flush_pkts()
        self.ble_sniffer.start_sniffer(self.rx_channel)
        return self.ble_sniffer

    def __exit__(self, exc_type, exc_value, traceback):
        if self.ble_sniffer:
            self.ble_sniffer.stop_sniffer()


import datetime
import time
from wiliot_api.api_client import WiliotCloudError
from wiliot_deployment_tools.api.extended_api import ExtendedEdgeClient
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.if_defines import DEFAULT_DELAY, SEP
from wiliot_deployment_tools.interface.pkt_generator import BrgPktGenerator
from wiliot_deployment_tools.gw_certificate.tests.generic import PASS_STATUS, GenericTest


class GenericBrgMgmtStage():
    def __init__(self, ble_sim:BLESimulator, edge:ExtendedEdgeClient, stage_name, **kwargs):
        #Clients
        self.ble_sim = ble_sim
        self.edge = edge
        #Stage Params
        self.stage_name = stage_name
        self.stage_pass = False
        self.report = ''
        self.start_time = None

    def prepare_stage(self):
        debug_print(f'### Starting Stage: {self.stage_name}')
        self.ble_sim.set_sim_mode(True)

    def add_to_stage_report(self, report):
        self.report += '\n' + report

class InitStage(GenericBrgMgmtStage):
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name='init_stage')
        self.pkt_gen = BrgPktGenerator()
        
    def run(self):
        self.start_time = datetime.datetime.now()
        cfg_sent = datetime.datetime.now()- datetime.timedelta(minutes=1)
        hb_sent = datetime.datetime.now()- datetime.timedelta(minutes=1)
        now = datetime.datetime.now()
        brg_claimed = False
        brg_id = self.pkt_gen.bridge_id
        print(brg_id)
        while datetime.datetime.now() - self.start_time < datetime.timedelta(minutes=5) and not brg_claimed:
            pkts = self.pkt_gen.get_new_packets()
            if now - datetime.timedelta(seconds=5) > hb_sent:
                brg_hb = pkts['brg_hb']
                self.ble_sim.send_packet(brg_hb)
                time.sleep(DEFAULT_DELAY)
                print(pkts['brg_hb'].__dict__)
                hb_sent = now
            if now - datetime.timedelta(seconds=5) > cfg_sent:
                brg_cfg = pkts['brg_cfg']
                self.ble_sim.send_packet(brg_cfg)
                time.sleep(DEFAULT_DELAY)
                print(pkts['brg_cfg'].__dict__)
                cfg_sent = now
            try:
                self.edge.claim_bridge(brg_id)
            except WiliotCloudError as e:
                print(e)
            now = datetime.datetime.now()

STAGES = [InitStage]

class BrgMgmtTest(GenericTest):
    def __init__(self, **kwargs):        
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, test_name=type(self).__name__)
        self.stages = [stage(**self.__dict__) for stage in STAGES]
    
    def run(self):
        self.start_time = datetime.datetime.now()
        self.test_pass = True
        for stage in self.stages:
            stage.prepare_stage()
            stage.run()
            # self.add_to_test_report(stage.generate_stage_report())
            # if stage.stage_pass == False:
            #     self.test_pass = False
        run_time = datetime.datetime.now() - self.start_time
        debug_print(f'\n{SEP}')
        debug_print(f'Bridge Management Test {PASS_STATUS[self.test_pass]}, Running time {run_time}')
        debug_print(self.report)



import setuptools
import sys
packages = setuptools.find_packages()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(name='wiliot-deployment-tools',
                 use_scm_version={
                     'git_describe_command': "git describe --long --tags --match [0-9]*.[0-9]*.[0-9]*",
                     'write_to': "wiliot_deployment_tools/version.py",
                     'write_to_template': '__version__ = "{version}"',
                     'root': ".",
                 },
                 setup_requires=['setuptools_scm'],
                 author='Wiliot',
                 author_email='support@wiliot.com',
                 description="A library for interacting with Wiliot's Deployment Tools",
                 long_description=long_description,
                 long_description_content_type="text/markdown",
                 url='',
                 project_urls={
                     "Bug Tracker": "https://WILIOT-ZENDESK-URL",
                 },
                 license='MIT',
                 classifiers=[
                     "Programming Language :: Python :: 3",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent",
                 ],
                 packages=packages,
                #  add all support files to the internal installation
                 package_data={"": ["*.*"]},
                 include_package_data=True,
                 install_requires=[
                     'wiliot-core>=4.0.9',
                     'wiliot-api>=4.4.0',
                     'setuptools_scm>=7.1.0',
                     'bitstruct==8.17.0',
                     'bokeh==2.4.1',
                     'boto3==1.26.77',
                     'colorama==0.4.6',
                     'jsonpickle==3.0.1',
                     'numpy>=1.21.5',
                     'pandas==1.4.2',
                     'plotly==5.13.0',
                     'pytz==2022.7.1',
                     'requests>=2.28.1',
                     'setuptools>=65.6.3',
                     'tabulate==0.8.10',
                     'backports.zoneinfo==0.2.1; python_version < "3.9.0"',
                     'tinytuya>=1.10.2',
                     'pyserial>=3.5',
                     'GitPython>=3.1.31',
                     'tzdata>=2023.3',
                     'confluent-kafka>=2.1.1',
                     'paho-mqtt>=1.6.1',
                     'packaging>=23.1',
                     'scipy>=1.11.1',
                     'mplcursors>=0.5.2',
                     'matplotlib>=3.7.2',
                     'pydantic',
                     'jinja2>=3.1.2',
                     'jsonschema>=4.19.0',
                     'statsmodels>=0.14.1'
                 ],
                 zip_safe=False,
                 python_requires='>=3.8',
                #  include_package_data=False,
                #  exclude_package_data={
                #      "": ['*.internal.*', '*.internal', 'internal', 'internal.*']},
                 entry_points={'console_scripts': [
                     'wlt-firmware=wiliot_deployment_tools.firmware_update.firmware_update_cli:main_cli',
                     'wlt-power-mgmt=wiliot_deployment_tools.power_mgmt.power_mgmt_cli:main_cli',
                     'wlt-calibration-mgmt=wiliot_deployment_tools.calibration_mgmt.calibration_mgmt_cli:main_cli',
                     'wlt-connectivity-analyzer=wiliot_deployment_tools.connectivity_analyzer.connectivity_analyzer_cli:main_cli',
                     'wlt-config=wiliot_deployment_tools.automatic_configuration_tool.configuration_tool_cli:main_cli',
                     'wlt-log=wiliot_deployment_tools.log_viewer.log_viewer_cli:main_cli',
                     'wlt-gw-certificate=wiliot_deployment_tools.gw_certificate.gw_certificate_cli:main_cli',
                     'wlt-kafka=wiliot_deployment_tools.internal.kafka_consumer.kafka_consumer_cli:main_cli',
                     'wlt-power-optimization=wiliot_deployment_tools.power_optimization_tool.current_gain_analysis_cli:main_cli',
                     'wlt-broker-change=wiliot_deployment_tools.interface.broker_change_cli:main_cli',]},
                 )



import datetime
import subprocess
import serial
import serial.tools.list_ports
import time
import os.path
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.if_defines import *
from packaging import version
import pkg_resources

LATEST_VERSION = '3.16.46'
LATEST_VERSION_FILE = f'{LATEST_VERSION}_app.zip'
LATEST_VERSION_PATH = pkg_resources.resource_filename(__name__, LATEST_VERSION_FILE)

class UARTError(Exception):
    pass

class UARTInterface:
    def __init__(self, comport, update_fw=True):
        self.comport = comport
        self.serial = serial.Serial(port=comport, baudrate=921600, timeout=SERIAL_TIMEOUT)
        self.serial.flushInput()
        self.gw_app_rx = None
        version_supported = self.check_fw_supported()
        if not version_supported and update_fw:
            update_status = self.update_firmware()
            if not update_status:
                raise UARTError('Update Failed! Update FW manually using NRF Tools')
        debug_print(f'Serial Connection {comport} Initialized')

    @staticmethod
    def get_comports():
        ports = serial.tools.list_ports.comports()
        debug_print(SEP + "\nAvailable ports:")
        for port, desc, hwid in sorted(ports):
            debug_print("{}: {} [{}]".format(port, desc, hwid))
        debug_print(SEP + "\n")
        return ports
    
    def read_line(self):
        # This reads a line from the ble device (from the serial connection using ble_ser),
        # strips it from white spaces and then decodes to string from bytes using the "utf-8" protocol.
        answer = self.serial.readline().strip().decode("utf-8", "ignore")
        if len(answer) == 0:
            return None
        return answer
    
    def write_ble_command(self, cmd, read=False): 
        # This function writes a command (cmd) to the ble using a serial connection (ble_ser) that are provided to it beforehand.. and returns the answer from the device as string
        debug_print("Write to BLE: {}".format(cmd)) 
        # Shows on terminal what command is about to be printed to the BLE device
        bytes_to_write = bytes(cmd.encode("utf-8")) + b'\r\n'
        self.serial.write(bytes_to_write)
        answer = None
        if read:
            # The "bytes" function converts the command from string to bytes by the specified "utf-8" protocol then we use .write to send the byte sequence to the ble device using the serial connection that we have for this port (ble_ser)
            # Pauses the program for execution for 0.01sec. This is done to allow the device to process the command and provide a response before reading the response.
            time.sleep(1)
            answer = self.read_line()
            debug_print(answer)
        return answer
    
    def flush(self):
        self.serial.close()
        self.serial.open()
        self.serial.flushInput()
        self.serial.flush()
        self.serial.readall()
    
    def reset_gw(self):
        self.flush()
        self.write_ble_command(RESET_GW)
        self.gw_app_rx = None
        time.sleep(3)
                    
    def cancel(self):
        self.write_ble_command(CANCEL)
        
    def set_rx(self, rx_channel):
        assert rx_channel in RX_CHANNELS
        if self.gw_app_rx is None:
            self.reset_gw()
        cmd = f'!gateway_app {rx_channel} 30 0 17'
        self.write_ble_command(cmd)
        self.gw_app_rx = rx_channel
        
    def set_sniffer(self, rx_channel):
        self.set_rx(rx_channel)
        self.flush()
        self.write_ble_command(SET_SNIFFER)
        self.flush()
    
    def cancel_sniffer(self):
        self.write_ble_command(CANCEL_SNIFFER)
        self.flush()
        
    def get_version(self):
        self.reset_gw()
        self.flush()
        timeout = datetime.datetime.now() + datetime.timedelta(seconds=15)
        while datetime.datetime.now() < timeout:
            raw_version = self.write_ble_command(VERSION, read=True)
            if raw_version is not None:
                if GW_APP_VERSION_HEADER in raw_version:
                    return raw_version.split(' ')[0].split('=')[1]
        return None
    
    def check_fw_supported(self):
        current_version = self.get_version()
        if current_version is None:
            raise UARTError("Cannot initialize board! Please try disconnecting and connecting USB cable")
        current_version = version.parse(current_version)
        hex_version = version.parse(os.path.splitext(os.path.basename(LATEST_VERSION_PATH))[0].split('_')[0])
        if current_version >= hex_version:
            debug_print(f'GW Running version {current_version}')
            return True
        return False
    
    def update_firmware(self):
        self.write_ble_command('!reset', read=True)
        self.write_ble_command('!move_to_bootloader', read=True)
        self.serial.close()
        p = subprocess.Popen(f'nrfutil dfu serial --package "{LATEST_VERSION_PATH}" -p {self.comport} -fc 0 -b 115200 -t 10', shell=True)
        p.wait()
        return_code = p.returncode
        debug_print('Waiting for device to update...')
        timeout = datetime.datetime.now() + datetime.timedelta(minutes=2)
        current_ver = ''
        self.serial.open()
        while GW_APP_VERSION_HEADER not in current_ver and datetime.datetime.now() < timeout:
            current_ver = self.write_ble_command(VERSION, read=True)
            if current_ver is None:
                current_ver = ''
            time.sleep(1)
        if current_ver.split(' ')[0].split('=')[1] != LATEST_VERSION:
            return False
        if return_code == 0:
            return True
        return False
            
        ### TODO: Sequence List


RX_CHANNELS = [37, 38, 39]
SET_SNIFFER = '!set_logger_mode 2'
CANCEL_SNIFFER = '!set_logger_mode 0'
CANCEL = '!cancel'
RESET_GW = '!reset'
VERSION = '!version'
SEND_ALL_ADV_CHANNELS = 0
SERIAL_TIMEOUT = 0.1
SEP = "#"*100
LENGTH = '1E'
GAP_TYPE = '16'
DEFAULT_ADVA = 'FFFFFFFFFFFF'
DEFAULT_DUPLICATES = 3
GW_APP_VERSION_HEADER = 'WILIOT_GW_BLE_CHIP_SW_VER'
DEFAULT_OUTPUT_POWER = 8
DEFAULT_DELAY = 15
COUPLING_DUPLICATIONS = [i for i in range(1, 6)]
COUPLING_TIME_DELAYS = [i for i in range(15, 256, 30)]
BRIDGES = [i for i in range(3)]
MAX_NFPKT = 65535
MAX_RSSI = 255
ADVA_LENGTH = 12
GAP_LENGTH = 4
SERVICE_UUID_LENGTH = 4
GROUP_ID_LENGTH = 6
UPLINK_DUPLICATIONS = [i for i in range(1, 6)]
UPLINK_TIME_DELAYS = [i for i in range(15, 256, 120)]

